import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    """

    dev = False

    # Create train, test from train split
    old_train = read_csv(raw / "train.csv")
    unique_image_ids = old_train["image_id"].unique()
    # Original train has 15k images, original test has 3k images
    # Our new train will have 13.5k images, our new test will have 1.5k images
    expected_train_size = 13500
    expected_test_size = 1500
    train_image_ids, test_image_ids = train_test_split(
        unique_image_ids, test_size=0.1, random_state=0
    )

    new_train = old_train[old_train["image_id"].isin(train_image_ids)]
    answers = old_train[old_train["image_id"].isin(test_image_ids)]

    # Create sample submission
    sample_submission = pd.DataFrame(
        {
            "image_id": test_image_ids,
            "PredictionString": "14 1 0 0 1 1",
        }  # As per the original sample submission
    )

    # Checks
    assert (
        len(set(new_train["image_id"])) == expected_train_size
    ), f"Expected {expected_train_size} train image_ids, got {len(set(new_train['image_id']))}"
    assert (
        len(set(answers["image_id"])) == expected_test_size
    ), f"Expected {expected_test_size} test image_ids, got {len(set(answers['image_id']))}"
    assert set(new_train["image_id"]).isdisjoint(
        set(answers["image_id"])
    ), f"image_id is not disjoint between train and test sets"
    assert (
        new_train.columns.tolist() == old_train.columns.tolist()
    ), f"Columns of new train and old train are not the same: {new_train.columns.tolist()} vs {old_train.columns.tolist()}"
    assert len(new_train) + len(answers) == len(
        old_train
    ), f"Length of new train and answers should add up to the length of old train, got {len(new_train) + len(answers)} vs {len(old_train)}"
    assert len(sample_submission) == len(
        set(answers["image_id"])
    ), f"Length of sample submission should be equal to the number of unique image_ids in answers, got {len(sample_submission)} vs {len(set(answers['image_id']))}"

    # Reformat answers
    def _get_consensus_annotation(answers, inspect_duplicates=False):
        """
        In the original train, there can be multiple annotations of the same image_id, class_id pair. (Different radiologists draw the bounding boxes differently for the same finding)

        In the original test, there is only one annotation per image_id, class_id pair. The original test set is labeled by consensus of 5 radiologists. (Source: https://www.kaggle.com/competitions/vinbigdata-chest-xray-abnormalities-detection/discussion/207969#1134645)

        We simulate consensus by taking the first annotation for each image_id, class_id pair.
        """

        if inspect_duplicates:
            duplicates = answers[answers.duplicated(subset=["image_id", "class_id"], keep=False)]
            duplicates = duplicates.sort_values(by=["image_id", "class_id"])
            duplicates.to_csv("duplicates.csv", index=False)

        answers = answers.groupby(by=["image_id", "class_id"]).first().reset_index()
        return answers

    answers = _get_consensus_annotation(answers)
    # Filling in missing values for when there is no finding (class_id = 14)
    answers = answers.fillna(0)
    answers.loc[answers["class_id"] == 14, ["x_max", "y_max"]] = 1.0

    # Create gold submission
    gold = answers[["image_id", "class_id", "x_min", "y_min", "x_max", "y_max"]].copy()
    # Create individual prediction strings
    gold.loc[:, "PredictionString"] = gold.apply(
        lambda row: f"{row['class_id']} 1.0 {row['x_min']} {row['y_min']} {row['x_max']} {row['y_max']}",
        axis=1,  # 1.0 is the confidence score
    )
    # Group by image_id and concatenate prediction strings
    gold = gold.groupby("image_id")["PredictionString"].agg(" ".join).reset_index()
    gold = gold.reset_index(drop=True)
    assert len(gold) == len(
        set(answers["image_id"])
    ), f"Length of gold should be equal to the number of unique image_ids in answers, got {len(gold)} vs {len(set(answers['image_id']))}"

    # Write CSVs
    new_train.to_csv(public / "train.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)
    answers.to_csv(private / "answers.csv", index=False)
    gold.to_csv(private / "gold_submission.csv", index=False)

    # Copy over files
    (public / "test").mkdir(exist_ok=True)
    (public / "train").mkdir(exist_ok=True)

    if dev == True:
        train_image_ids = train_image_ids[:10]
        test_image_ids = test_image_ids[:10]

    for file_id in tqdm(train_image_ids, desc="Copying train files"):
        shutil.copyfile(
            src=raw / "train" / f"{file_id}.dicom",
            dst=public / "train" / f"{file_id}.dicom",
        )

    for file_id in tqdm(test_image_ids, desc="Copying test files"):
        shutil.copyfile(
            src=raw / "train" / f"{file_id}.dicom",
            dst=public / "test" / f"{file_id}.dicom",
        )

    # Check files
    assert len(list(public.glob("train/*.dicom"))) == len(
        train_image_ids
    ), f"Expected {len(train_image_ids)} train files, got {len(list(public.glob('train/*.dicom')))}"
    assert len(list(public.glob("test/*.dicom"))) == len(
        test_image_ids
    ), f"Expected {len(test_image_ids)} test files, got {len(list(public.glob('test/*.dicom')))}"



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

from mlebench.competitions.utils import df_to_one_hot
from mlebench.utils import read_csv

from .dogs import DOGS_LIST


def to_one_hot(df: pd.DataFrame) -> pd.DataFrame:
    return df_to_one_hot(df, id_column="id", target_column="breed", classes=DOGS_LIST)


def _save_split(
    raw_data_path: Path,
    public_dir: Path,
    private_dir: Path,
    train_df: pd.DataFrame,
    test_df: pd.DataFrame,
):
    """Saves a given train/test split to the specified public and private directories."""
    # Ensure target directories exist
    public_dir.mkdir(exist_ok=True)
    private_dir.mkdir(exist_ok=True)

    # one hot the private test because we will one-hot the submission, as per kaggle.com
    test_labels_private = to_one_hot(test_df.copy())

    # Copy over files
    train_df.to_csv(public_dir / "labels.csv", index=False)
    test_labels_private.to_csv(private_dir / "test.csv", index=False)

    (public_dir / "train").mkdir(exist_ok=True)
    for file_id in train_df["id"]:
        shutil.copyfile(
            src=raw_data_path / "train" / f"{file_id}.jpg",
            dst=public_dir / "train" / f"{file_id}.jpg",
        )

    (public_dir / "test").mkdir(exist_ok=True)
    for file_id in test_df["id"]:
        shutil.copyfile(
            src=raw_data_path / "train" / f"{file_id}.jpg",
            dst=public_dir / "test" / f"{file_id}.jpg",
        )

    # Check integrity of the files copied
    assert len(list(public_dir.glob("train/*.jpg"))) == len(train_df)
    assert len(list(public_dir.glob("test/*.jpg"))) == len(test_df)

    # Create a sample submission file
    submission_df = test_labels_private.copy()
    for col in submission_df.columns[1:]:
        submission_df[col] = submission_df[col].astype("float64")
    submission_df.iloc[:, 1:] = 1 / 120
    submission_df.to_csv(public_dir / "sample_submission.csv", index=False)

    assert submission_df.shape == (len(test_df), 121)  # 1 id column + 120 breeds


def prepare(raw: Path, public: Path, private: Path):
    # Read the full raw dataset labels
    all_labels = read_csv(raw / "labels.csv")

    # --- 1. Original Split: Create main train and test sets ---
    # This split and its outputs must remain identical to the original script.
    train_df, test_df = train_test_split(all_labels, test_size=0.1, random_state=0)

    # Save the original split to the 'public' and 'private' directories
    _save_split(
        raw_data_path=raw,
        public_dir=public,
        private_dir=private,
        train_df=train_df,
        test_df=test_df,
    )

    # --- 2. New Split: Create a validation set from the main train set ---
    # Define paths for the new validation directories, parallel to the original ones.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # To ensure the new validation set ('test_val_df') is the same size as the
    # original test set, we calculate the required test_size for the second split.
    test_size_for_val_split = len(test_df) / len(train_df)

    # Split the main training data (train_df) into a smaller training set
    # and a validation set, using the same random_state for reproducibility.
    train_val_df, test_val_df = train_test_split(
        train_df, test_size=test_size_for_val_split, random_state=0
    )

    # Save the new validation split to the 'public_val' and 'private_val' directories
    # using the same helper function to ensure identical structure and filenames.
    _save_split(
        raw_data_path=raw,
        public_dir=public_val,
        private_dir=private_val,
        train_df=train_val_df,
        test_df=test_val_df,
    )


import shutil
from pathlib import Path

from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path):

    # Create train, test from train split
    old_train = read_csv(raw / "train.csv")
    new_train, answers = train_test_split(old_train, test_size=0.1, random_state=0)

    # Create sample submission
    submission_df = answers.copy()
    submission_df["time_to_eruption"] = 0

    # Checks
    assert len(answers) == len(submission_df), "Answers and submission should have the same length"
    assert not any(
        new_train["segment_id"].isin(answers["segment_id"])
    ), "No segment_id should be in both train and answers"
    assert list(new_train.columns) == [
        "segment_id",
        "time_to_eruption",
    ], "new_train should have columns 'segment_id' and 'time_to_eruption'"
    assert list(submission_df.columns) == [
        "segment_id",
        "time_to_eruption",
    ], "submission_df should have columns 'segment_id' and 'time_to_eruption'"
    assert list(answers.columns) == [
        "segment_id",
        "time_to_eruption",
    ], "answers should have columns 'segment_id' and 'time_to_eruption'"
    assert len(new_train) + len(answers) == len(
        old_train
    ), "The sum of the length of new_train and answers should be equal to the length of old_train"

    # Write CSVs
    answers.to_csv(private / "test.csv", index=False)
    new_train.to_csv(public / "train.csv", index=False)
    submission_df.to_csv(public / "sample_submission.csv", index=False)

    # Copy over files
    (public / "test").mkdir(exist_ok=True)
    (public / "train").mkdir(exist_ok=True)

    for file_id in tqdm(new_train["segment_id"], desc="Copying train files"):
        shutil.copyfile(
            src=raw / "train" / f"{file_id}.csv",
            dst=public / "train" / f"{file_id}.csv",
        )

    for file_id in tqdm(answers["segment_id"], desc="Copying test files"):
        shutil.copyfile(
            src=raw / "train" / f"{file_id}.csv",
            dst=public / "test" / f"{file_id}.csv",
        )

    # Checks on files
    assert len(list(public.glob("train/*.csv"))) == len(
        new_train
    ), "Public train should have the same number of files as new_train"
    assert len(list(public.glob("test/*.csv"))) == len(
        answers
    ), "Public test should have the same number of files as answers"



from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def _create_split_files(
    train_df: pd.DataFrame,
    test_df: pd.DataFrame,
    public_dir: Path,
    private_dir: Path,
):
    """A helper function to save train/test splits and artifacts."""
    # Create random example submission
    example_submission = test_df[["PassengerId", "Transported"]].copy()
    example_submission["Transported"] = False
    example_submission.to_csv(public_dir / "sample_submission.csv", index=False)

    # Create private files
    test_df.to_csv(private_dir / "test.csv", index=False)

    # Create public files visible to agents
    train_df.to_csv(public_dir / "train.csv", index=False)
    test_df.drop("Transported", axis="columns").to_csv(
        public_dir / "test.csv", index=False
    )


def prepare(raw: Path, public: Path, private: Path):
    # Define and create all output directories to ensure they exist
    public.mkdir(parents=True, exist_ok=True)
    private.mkdir(parents=True, exist_ok=True)
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"
    public_val.mkdir(parents=True, exist_ok=True)
    private_val.mkdir(parents=True, exist_ok=True)

    # --- Original Split: Create main train and test sets ---
    # Create train and test splits from train set
    old_train = read_csv(raw / "train.csv")
    new_train, new_test = train_test_split(old_train, test_size=0.1, random_state=0)

    # Generate the files for the original public/private directories
    # This ensures the original outputs are unchanged
    _create_split_files(new_train, new_test, public, private)

    # --- Validation Split: Create a new train and validation set ---
    # The new validation set should be approx. the same size as the original test set.
    # We calculate the required test_size for splitting the new_train set.
    test_size_for_val = len(new_test) / len(new_train)

    # Split the training data again to create a new, smaller training set and a validation set
    train_val, test_val = train_test_split(
        new_train, test_size=test_size_for_val, random_state=0
    )

    # Generate the files for the new validation directories, using the exact same
    # structure and filenames as the original split.
    _create_split_files(train_val, test_val, public_val, private_val)


from abc import ABC, abstractmethod
from pathlib import Path
import tempfile
import logging
from typing import Tuple, Any
from dsat.models.task import TaskDefinition
from dsat.services.data_analyzer import DataAnalyzer

logger = logging.getLogger(__name__)


class TaskHandler(ABC):
    """
    Base class for handlers that translate between logical TaskDefinition and physical file interfaces required by DSATWorkflow.

    Each handler encapsulates preparation and parsing logic for specific task types (e.g., Kaggle, QA),
    allowing the workflow itself to remain task-agnostic.
    """
    def __init__(self):
        """
        Initialize the handler and create a temporary, self-managed directory
        for storing physical files generated for the task.
        """
        try:
            self.temp_dir = tempfile.TemporaryDirectory()
        except Exception as e:
            logger.error(f"Failed to create temporary directory for TaskHandler: {e}")
            self.temp_dir = None
        
        self.analyzer = DataAnalyzer()

    @abstractmethod
    def prepare_input(self, task: TaskDefinition) -> Tuple[str, str, Path, Path]:
        """
        Prepare the physical file input required by the workflow.

        This method converts logical tasks into physical parameters needed by DSATWorkflow.solve().

        Args:
            task: Logical task definition.

        Returns:
            A tuple (description, io_instructions, data_dir, output_path) to pass to workflow.solve().
        """
        raise NotImplementedError

    @abstractmethod
    def parse_output(self, output_path: Path) -> Any:
        """
        Parse the workflow's output file into structured results required by benchmarking.

        This method converts physical output files back into logical answers.

        Args:
            output_path: Path where the workflow saved its output.

        Returns:
            Final answer in the format expected by benchmarking (e.g., string for QA, Path object for Kaggle).
        """
        raise NotImplementedError

    def cleanup(self):
        """
        Explicitly clean up the temporary directory.
        """
        if self.temp_dir:
            try:
                self.temp_dir.cleanup()
                logger.debug(f"Successfully cleaned up temporary directory for {self.__class__.__name__}.")
            except Exception as e:
                logger.error(f"Error cleaning up temporary directory for {self.__class__.__name__}: {e}")

    def __del__(self):
        """Ensure cleanup is called when the object is garbage collected."""
        self.cleanup()


class KaggleTaskHandler(TaskHandler):
    """
    Handler for Kaggle-style file input/file output tasks.
    This is a "pass-through" implementation since tasks are already file-based.
    """
    def prepare_input(self, task: TaskDefinition) -> Tuple[str, str, Path, Path]:
        """Extract paths, analyze data, and separate description from I/O instructions."""
        if task.task_type != "kaggle":
            raise ValueError("KaggleTaskHandler can only handle tasks of type 'kaggle'.")

        description = task.payload.get("description")
        data_dir = Path(task.payload.get("public_data_dir"))
        output_path = Path(task.payload.get("output_submission_path"))

        if not all([description, data_dir, output_path]):
            raise ValueError("Kaggle task payload is missing required keys: 'description', 'public_data_dir', 'output_submission_path'.")
        if not data_dir.exists() or not data_dir.is_dir():
            raise FileNotFoundError(f"Kaggle public_data_dir not found: {data_dir}")

        logger.info(f"Analyzing input data for task '{task.task_id}'...")
        
        data_report = self.analyzer.analyze_data(data_dir, task_type="kaggle")
        io_instructions = self.analyzer.generate_io_instructions(output_path.name, optimization_context=False)

        augmented_description = f"{description}\n{data_report}"

        logger.debug(f"Preparing Kaggle task '{task.task_id}': data_dir='{data_dir}', output_path='{output_path}'")
        return augmented_description, io_instructions, data_dir, output_path

    def parse_output(self, output_path: Path) -> Path:
        """
        For Kaggle tasks, the result is the output file itself.
        This just validates that the file was created.
        """
        if not output_path.exists():
            # In actual evaluation, this will be caught and reported as a failure.
            logger.warning(f"Agent did not produce the required submission file at: {output_path}")
            # Return the path even if it doesn't exist, let the caller (e.g., benchmark) handle the file not found case.
        return output_path


class QATaskHandler(TaskHandler):
    """
    Handler for simple question-answer (QA) tasks.
    This is a "translation" implementation that converts string questions to files and expects answers as files.
    """
    def prepare_input(self, task: TaskDefinition) -> Tuple[str, str, Path, Path]:
        """Convert QA question to physical file input."""
        if task.task_type != "qa":
            raise ValueError("QATaskHandler can only handle tasks of type 'qa'.")
        if not self.temp_dir:
            raise RuntimeError("Temporary directory not available for QATaskHandler.")

        question = task.payload.get("question")
        if not question:
            raise ValueError("QA task payload is missing required key: 'question'.")

        data_dir = Path(self.temp_dir.name)
        
        # Create physical task representation
        problem_file = data_dir / "problem.txt"
        problem_file.write_text(question, encoding='utf-8')
        
        # Define output contract
        output_path = data_dir / "answer.txt"
        
        # This core instruction is now simpler
        core_instruction = (
            "Your task is to answer the question found in `problem.txt`. "
            "Write ONLY the final answer into the required output file."
        )

        data_report = self.analyzer.analyze_data(data_dir, task_type="qa")
        io_instructions = self.analyzer.generate_io_instructions(output_path.name, optimization_context=False)
        
        description = f"{core_instruction}\n{data_report}"
        
        logger.debug(f"Preparing QA task '{task.task_id}': input file='{problem_file}', expected output='{output_path}'")
        return description, io_instructions, data_dir, output_path

    def parse_output(self, output_path: Path) -> str:
        """Read and return the final answer string from the output file."""
        if not output_path.exists() or not output_path.is_file():
            logger.warning(f"Agent did not produce the answer file for QA task at: {output_path}")
            return "[ERROR] Agent did not produce an answer file."
        
        try:
            answer = output_path.read_text(encoding='utf-8').strip()
            logger.debug(f"Parsed QA answer from '{output_path}': '{answer[:50]}...'")
            return answer
        except Exception as e:
            logger.error(f"Failed to read or parse QA answer file '{output_path}': {e}")
            return f"[ERROR] Failed to parse answer file: {e}"


class DataSciTaskHandler(TaskHandler):
    """
    Handler for DataSciBench tasks.
    These are multi-step data science tasks with prompts and optional input files.
    """
    def prepare_input(self, task: TaskDefinition) -> Tuple[str, str, Path, Path]:
        """Prepare DataSciBench task input."""
        if task.task_type != "datasci":
            raise ValueError("DataSciTaskHandler can only handle tasks of type 'datasci'.")

        prompt = task.payload.get("prompt", "")
        input_dir = task.payload.get("input_dir", "")
        output_dir = task.payload.get("output_dir", "")

        if not prompt:
            raise ValueError("DataSci task payload is missing required key: 'prompt'.")

        # Use input_dir as data_dir, or temp_dir if no input files
        if input_dir and Path(input_dir).exists():
            data_dir = Path(input_dir)
        elif self.temp_dir:
            data_dir = Path(self.temp_dir.name)
        else:
            raise RuntimeError("No data directory available for DataSciTaskHandler.")

        # Output directory
        if output_dir:
            output_path = Path(output_dir) / "output.csv"
        else:
            output_path = data_dir / "output.csv"

        # Build description with the prompt
        description = prompt

        # Analyze data if available
        try:
            data_report = self.analyzer.analyze_data(data_dir, task_type="datasci")
            description = f"{prompt}\n\n{data_report}"
        except Exception as e:
            logger.debug(f"Data analysis skipped: {e}")

        # Generate I/O instructions
        io_instructions = (
            f"All input data files are in the current working directory.\n"
            f"Save all output files to the current working directory.\n"
            f"Follow the task instructions carefully and generate the required output files."
        )

        logger.debug(f"Preparing DataSci task '{task.task_id}': data_dir='{data_dir}', output_dir='{output_dir}'")
        return description, io_instructions, data_dir, output_path

    def parse_output(self, output_path: Path) -> Path:
        """
        For DataSci tasks, return the output directory path.
        The actual evaluation is done by the benchmark class using metric.yaml.
        """
        if output_path.parent.exists():
            return output_path.parent
        return output_path


class OpenEndedTaskHandler(TaskHandler):
    """
    Handler for open-ended tasks (mathematical modeling, simulations, strategy tasks).
    These tasks don't have ground truth answers and are evaluated via LLM judges.
    """
    def prepare_input(self, task: TaskDefinition) -> Tuple[str, str, Path, Path]:
        """Prepare open-ended task input."""
        if task.task_type != "open_ended":
            raise ValueError("OpenEndedTaskHandler can only handle tasks of type 'open_ended'.")
        if not self.temp_dir:
            raise RuntimeError("Temporary directory not available for OpenEndedTaskHandler.")

        # Get task paths from payload
        raw_dir_str = task.payload.get("raw_data_dir", "")
        description_file = task.payload.get("description_file", "")
        rubric_file = task.payload.get("rubric_file", "")

        # Use temp directory as working directory
        data_dir = Path(self.temp_dir.name)

        # Copy ONLY data files (CSV, JSON, etc.) - exclude description and rubric files
        if raw_dir_str:
            raw_dir = Path(raw_dir_str)
            if raw_dir.exists():
                import shutil
                for file in raw_dir.iterdir():
                    if file.is_file() and file.suffix in ['.csv', '.json', '.txt', '.xlsx', '.parquet']:
                        # Exclude description.md and rubric.md from being treated as data files
                        if file.name not in ['description.md', 'rubric.md']:
                            shutil.copy2(file, data_dir / file.name)
                            logger.debug(f"Copied data file: {file.name}")

        # Read task description and rubric from files if provided
        description = task.payload.get("description", "")
        rubric = task.payload.get("rubric", "")

        # Read from files if specified
        if description_file and Path(description_file).exists():
            try:
                description = Path(description_file).read_text(encoding='utf-8')
                logger.debug(f"Read description from {description_file} ({len(description)} chars)")
            except Exception as e:
                logger.warning(f"Failed to read description file {description_file}: {e}")

        if rubric_file and Path(rubric_file).exists():
            try:
                rubric = Path(rubric_file).read_text(encoding='utf-8')
                logger.debug(f"Read rubric from {rubric_file} ({len(rubric)} chars)")
            except Exception as e:
                logger.warning(f"Failed to read rubric file {rubric_file}: {e}")

        if not description:
            raise ValueError("Open-ended task payload is missing required key: 'description'.")

        # Output path - agent should create an artifacts directory or report
        output_path = data_dir / "artifacts"

        # Build the FULL task description directly in the prompt
        # Include description and evaluation criteria
        # IMPORTANT: For open-ended tasks, explicitly require artifacts directory creation

        task_description_section = f"""## Task Description

{description}
"""

        if rubric:
            task_description_section += f"""

## Evaluation Criteria

{rubric}
"""

        # Analyze available data files to provide schema information (excluding task files)
        data_report = self.analyzer.analyze_data(data_dir, task_type="datasci")

        # Combine everything into the full description
        # Note: data_report already contains "--- COMPREHENSIVE DATA REPORT ---" header
        full_description = f"""{task_description_section}

{data_report}

## CRITICAL OUTPUT INSTRUCTIONS

**YOU MUST CREATE AN `artifacts/` DIRECTORY AND SAVE ALL OUTPUTS THERE:**

```python
import os
artifact_dir = 'artifacts'
os.makedirs(artifact_dir, exist_ok=True)

# Save all your work to the artifacts directory:
# - Analysis code: artifacts/analysis.py
# - Visualizations: artifacts/plot_*.png
# - Data files: artifacts/results.csv
# - Models, notebooks, etc.
```

## Task Goals
- Your goal is to complete this task to the best of your ability
- Create appropriate output files (code, analysis, visualizations, etc.) in the `artifacts/` subdirectory
- The evaluation will be based on the quality and completeness of your work according to the evaluation criteria
"""

        # Generate I/O instructions - VERY EXPLICIT for open-ended tasks
        io_instructions = f"""**OUTPUT DIRECTORY STRUCTURE (MANDATORY):**

```python
# At the START of your code, create the artifacts directory:
import os
artifact_dir = 'artifacts'
os.makedirs(artifact_dir, exist_ok=True)

# Save ALL outputs to this directory:
# - Code: f"{{artifact_dir}}/solution.py"
# - Plots: f"{{artifact_dir}}/visualization_{{i}}.png"
# - Data: f"{{artifact_dir}}/results.csv"
```

**REQUIREMENTS:**
1. Create the `artifacts/` directory at the beginning of your code
2. Save ALL generated files (plots, models, data, code) to this directory
3. Do NOT save files to the current directory - use the artifacts/ subdirectory
4. Focus on quality, completeness, and following the evaluation criteria
"""

        logger.debug(f"Preparing open-ended task '{task.task_id}': output_path='{output_path}', description_len={len(description)}")
        return full_description, io_instructions, data_dir, output_path

    def parse_output(self, output_path: Path) -> Path:
        """
        For open-ended tasks, return the artifacts directory path.
        The actual evaluation is done by LLM judges, not CSV grading.
        """
        if not output_path.exists():
            # If artifacts directory doesn't exist, return the parent temp dir
            # This allows evaluation to proceed even if no artifacts were created
            logger.warning(f"Open-ended task did not create artifacts directory at: {output_path}")
            return output_path.parent

        logger.debug(f"Parsed open-ended task artifacts from: {output_path}")
        return output_path


import logging
from typing import Dict

from dsat.models.formats import ComplexityScore, DecomposedPlan
from dsat.operators.base import Operator
from dsat.services.llm import LLMService

logger = logging.getLogger(__name__)

class ComplexityScorerOperator(Operator):
    """Scores the complexity of a natural language plan."""
    async def __call__(self, plan: str, task_goal: str) -> ComplexityScore:
        logger.info("Scoring plan complexity...")
        prompt = (
            "You are an expert project manager. On a scale of 1 to 5, where 1 is trivial "
            "(e.g., load a file and report basic properties) and 5 is highly complex (e.g., multi-stage pipeline involving custom algorithms, extensive preprocessing, or complex simulations), how complex is the following plan?\n\n"
            f"## Task Goal:\n{task_goal}\n\n"
            f"## Proposed Plan:\n{plan}\n\n"
            "Respond with a JSON object containing your score and a brief justification."
        )
        score = await self.llm_service.call_with_json(prompt, output_model=ComplexityScore)
        logger.info(f"Plan complexity scored at {score.complexity}/5.")
        return score

class PlanDecomposerOperator(Operator):
    """Decomposes a complex plan into a structured list of sequential steps."""
    async def __call__(self, plan: str, task_goal: str) -> DecomposedPlan:
        logger.info("Decomposing complex plan into steps...")
        prompt = (
            "You are an expert data scientist. Decompose the following high-level plan into a sequence of "
            "small, logical, and executable steps. Each step should represent a single cell in a data science notebook.\n\n"
            f"## Task Goal:\n{task_goal}\n\n"
            f"## High-Level Plan:\n{plan}\n\n"
            "Respond with a JSON object containing a list of tasks. Each task must have a unique `task_id` and an `instruction`."
        )
        decomposed_plan = await self.llm_service.call_with_json(prompt, output_model=DecomposedPlan)
        logger.info(f"Plan decomposed into {len(decomposed_plan.tasks)} steps.")
        return decomposed_plan



import os
import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import extract, read_csv


def create_dataframe_from_directory(directory: str) -> pd.DataFrame:
    """
    Creates a DataFrame from a directory of images.

    Args:
        directory (str): The path to the directory containing subdirectories of images.

    Returns:
        pd.DataFrame: A DataFrame with two columns: 'image' and 'label'. The 'image' column contains the file paths to the images, and the 'label' column contains the corresponding labels (subdirectory names).
    """
    data = []
    for label in sorted(os.listdir(directory)):  # Sort labels for determinism
        label_path = os.path.join(directory, label)
        if os.path.isdir(label_path):
            for file_name in sorted(os.listdir(label_path)):  # Sort files for determinism
                if file_name.endswith(".png"):
                    file_path = os.path.join(label_path, file_name)
                    data.append({"file": os.path.basename(file_path), "species": label})
    return pd.DataFrame(data)


def _process_split(
    train_df: pd.DataFrame,
    test_df: pd.DataFrame,
    public_path: Path,
    private_path: Path,
    source_images_dir: Path,
):
    """
    Helper function to process a train/test split and write all necessary files and folders.
    This function creates the public and private directories, generates CSVs, and copies image files.
    """
    # Ensure destination directories exist
    public_path.mkdir(exist_ok=True)
    private_path.mkdir(exist_ok=True)

    # Create a sample submission file
    submission_df = test_df.copy()
    submission_df["species"] = "Sugar beet"

    # Checks
    assert len(test_df) == len(submission_df), "Answers and submission should have the same length"
    assert not set(train_df["file"]).intersection(
        set(test_df["file"])
    ), "new_train and answers should not share any image"
    assert (
        "file" in train_df.columns and "species" in train_df.columns
    ), "Train DataFrame must have 'file' and 'species' columns"
    assert (
        "file" in submission_df.columns and "species" in submission_df.columns
    ), "Sample submission DataFrame must have 'file' and 'species' columns"

    # Write CSVs
    test_df.to_csv(private_path / "answers.csv", index=False)
    submission_df.to_csv(public_path / "sample_submission.csv", index=False)

    # Prepare image directories
    public_test_images_path = public_path / "test"
    public_train_images_path = public_path / "train"
    public_test_images_path.mkdir(exist_ok=True)
    public_train_images_path.mkdir(exist_ok=True)

    # Create nested folder structure for train
    for species in train_df["species"].unique():
        (public_train_images_path / species).mkdir(parents=True, exist_ok=True)

    # Use public path name for progress bar description
    desc_prefix = public_path.name.capitalize()

    for _, row in tqdm(
        train_df.iterrows(), desc=f"Copying {desc_prefix} Train Images", total=len(train_df)
    ):
        src_path = source_images_dir / row["species"] / row["file"]
        dst_path = public_train_images_path / row["species"] / row["file"]
        shutil.copyfile(src=src_path, dst=dst_path)

    for _, row in tqdm(
        test_df.iterrows(), desc=f"Copying {desc_prefix} Test Images", total=len(test_df)
    ):
        src_path = source_images_dir / row["species"] / row["file"]
        dst_path = public_test_images_path / row["file"]
        shutil.copyfile(src=src_path, dst=dst_path)

    # Final checks on copied files
    assert len(list(public_train_images_path.glob("**/*.png"))) == len(
        train_df
    ), f"Public train images should have the same number of images as the train DataFrame: number of files {len(list(public_train_images_path.glob('**/*.png')))} != len(train_df)={len(train_df)}"
    assert len(list(public_test_images_path.glob("*.png"))) == len(
        test_df
    ), f"Public test images should have the same number of images as the answers DataFrame: number of files {len(list(public_test_images_path.glob('*.png')))} != len(test_df)={len(test_df)}"


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    It then creates a secondary validation split (public_val, private_val) from the main training data.
    """
    # Directory containing the original images
    train_dir = raw / "train"
    old_train = create_dataframe_from_directory(train_dir)
    test_ratio = 0.14  # 794/(4750+794) = 0.14

    # --- 1. Original Split: Create main train and test sets ---
    train_df, test_df = train_test_split(old_train, test_size=test_ratio, random_state=0)

    # Check for original split integrity
    assert len(train_df) + len(test_df) == len(
        old_train
    ), "The combined length of new_train and answers should equal the length of old_train"

    # Process and save the original split to public/ and private/
    # This generates the original, unmodified competition output.
    _process_split(
        train_df=train_df,
        test_df=test_df,
        public_path=public,
        private_path=private,
        source_images_dir=train_dir,
    )

    # --- 2. New Validation Split: Split the main train_df again ---
    # Define paths for the new validation dataset, parallel to the original ones.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Calculate the new test ratio to make the validation set (test_val_df)
    # approximately the same size as the original test set (test_df).
    new_test_ratio = len(test_df) / len(train_df)

    # Split the training data to create a new, smaller training set and a validation set.
    # We use the same random_state for reproducibility and consistency with the first split.
    train_val_df, test_val_df = train_test_split(train_df, test_size=new_test_ratio, random_state=0)

    # Check for new split integrity
    assert (
        len(train_val_df) + len(test_val_df) == len(train_df)
    ), "The combined length of train_val_df and test_val_df should equal the length of train_df"

    # Process and save the new validation split to public_val/ and private_val/
    # Using the same helper ensures an identical file and directory structure.
    _process_split(
        train_df=train_val_df,
        test_df=test_val_df,
        public_path=public_val,
        private_path=private_val,
        source_images_dir=train_dir,
    )


from pathlib import Path

from sklearn.model_selection import train_test_split

from mlebench.utils import extract, read_csv

from .classes import CLASSES


def prepare(raw: Path, public: Path, private: Path):
    # extract only what is needed
    extract(raw / "train.csv.zip", raw)
    extract(raw / "test.csv.zip", raw)
    extract(raw / "test_labels.csv.zip", raw)
    extract(raw / "sample_submission.csv.zip", raw)

    # the test set is provided, so we dont have to split the train set nor form the sample submission
    train_with_labels = read_csv(raw / "train.csv")
    test_without_labels = read_csv(raw / "test.csv")
    answers = read_csv(raw / "test_labels.csv")
    sample_submission = read_csv(raw / "sample_submission.csv")
    sample_submission[CLASSES] = 0.5

    # save to public
    train_with_labels.to_csv(public / "train.csv", index=False)
    test_without_labels.to_csv(public / "test.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    # save to private
    answers.to_csv(private / "test.csv", index=False)

    assert len(answers) == len(
        sample_submission
    ), "Private test set and sample submission should be of the same length"

    assert sorted(answers["id"]) == sorted(
        test_without_labels["id"]
    ), "Private and Public test IDs should match"
    assert sorted(sample_submission["id"]) == sorted(
        test_without_labels["id"]
    ), "Public test and sample submission IDs should match"
    assert (
        len(set(train_with_labels["id"]) & set(test_without_labels["id"])) == 0
    ), "Train and test IDs should not overlap"

    # ==================================================================
    # === New code for creating the validation set starts here ===
    # The code above this line is untouched to ensure original outputs
    # remain identical.
    # ==================================================================

    # 1. Define new paths and create the directories for the validation set.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"
    public_val.mkdir(exist_ok=True)
    private_val.mkdir(exist_ok=True)

    # 2. Split the original training data to create a new, smaller training set
    # and a new validation set. The size of the validation set will be the same
    # as the size of the original test set.
    validation_set_size = len(test_without_labels)
    train_val, test_val_with_labels = train_test_split(
        train_with_labels,
        test_size=validation_set_size,
        random_state=42,  # Use a fixed random state for reproducibility
    )

    # 3. Prepare the validation set files, mirroring the original test set structure.
    # The public part (input features, without labels)
    test_val_without_labels = test_val_with_labels[["id", "comment_text"]].copy()

    # The private part (ground truth labels for the validation set)
    answers_val = test_val_with_labels[["id"] + CLASSES].copy()

    # 4. Create a sample submission file for the new validation set,
    # mirroring the original sample submission format.
    sample_submission_val = test_val_without_labels[["id"]].copy()
    sample_submission_val[CLASSES] = 0.5

    # 5. Save the new sets to the 'public_val' and 'private_val' directories,
    # using the same filenames as in the original 'public' and 'private' dirs.
    # Save to public_val
    train_val.to_csv(public_val / "train.csv", index=False)
    test_val_without_labels.to_csv(public_val / "test.csv", index=False)
    sample_submission_val.to_csv(public_val / "sample_submission.csv", index=False)

    # Save to private_val
    answers_val.to_csv(private_val / "test.csv", index=False)


"""Data preparation for ScienceBench task 101 (experimental band-gap prediction)."""

from __future__ import annotations

import shutil
from pathlib import Path

import pandas as pd


DATASET_NAME = "experimental_band_gap"
PRED_FILENAME = "experimental_band_gap_prediction_pred.csv"
GOLD_FILENAME = "experimental_band_gap_prediction_gold.csv"
SAMPLE_FILENAME = "sample_submission.csv"


def _repo_root() -> Path:
    return Path(__file__).resolve().parents[2]


def _dataset_dir() -> Path:
    return _repo_root() / "benchmark" / "datasets" / DATASET_NAME


def _gold_path() -> Path:
    return (
        _repo_root()
        / "benchmark"
        / "eval_programs"
        / "gold_results"
        / GOLD_FILENAME
    )


def _ensure_dir(path: Path) -> None:
    path.mkdir(parents=True, exist_ok=True)


def _copy_dataset(src: Path, public: Path) -> None:
    dest_root = public / DATASET_NAME
    dest_root.mkdir(parents=True, exist_ok=True)
    copied = 0
    for item in src.iterdir():
        if not item.is_file():
            continue
        shutil.copy2(item, dest_root / item.name)
        copied += 1
    print(f"✓ Copied {copied} dataset file(s) to {dest_root}")


def prepare(raw: Path, public: Path, private: Path) -> None:
    print("=" * 60)
    print("Preparing ScienceBench Task 101")
    print("Dataset:", DATASET_NAME)
    print("=" * 60)
    print("Raw directory:", raw)
    print("Public directory:", public)
    print("Private directory:", private)

    source_dir = raw if raw.exists() else _dataset_dir()
    if not source_dir.exists():
        raise FileNotFoundError(f"Dataset directory not found: {source_dir}")

    gold_path = _gold_path()
    if not gold_path.exists():
        raise FileNotFoundError(f"Gold CSV not found: {gold_path}")

    _ensure_dir(public)
    _ensure_dir(private)

    _copy_dataset(source_dir, public)

    gold_df = pd.read_csv(gold_path)
    sample = gold_df.head(3).copy()
    sample["gap_expt_eV"] = 0.0
    sample.to_csv(public / SAMPLE_FILENAME, index=False)
    print("✓ Created sample_submission.csv placeholder")

    gold_df.to_csv(private / "answer.csv", index=False)
    print("✓ Copied gold CSV to private directory")

    (private / "notes.txt").write_text(
        f"Expected submission: pred_results/{PRED_FILENAME}\nMAE threshold: 0.6 eV\n",
        encoding="utf-8",
    )
    print("✓ Wrote notes.txt")

    print("Data preparation completed.")



from pathlib import Path
import pandas as pd


def prepare(raw: Path, public: Path, private: Path):
    """
    Prepare the DABench task 309 dataset.

    Args:
        raw: Path to raw data directory (should contain titanic.csv)
        public: Path to public directory (for participants)
        private: Path to private directory (for grading)
    """
    # Load the data
    data_file = raw / "titanic.csv"
    if not data_file.exists():
        raise FileNotFoundError(f"Data file not found: {data_file}")

    df = pd.read_csv(data_file)

    # Save the full dataset to public directory
    train_file = public / "train.csv"
    df.to_csv(train_file, index=False)
    print(f"Saved training data to {train_file} ({len(df)} rows)")

    # Create sample submission file
    sample_submission = pd.DataFrame({
        'id': [309],
        'answer': ['@placeholder[0.00]']  # Placeholder answer
    })
    sample_submission_file = public / "sample_submission.csv"
    sample_submission.to_csv(sample_submission_file, index=False)
    print(f"Created sample submission: {sample_submission_file}")

    # Create answer file (ground truth)
    answer = pd.DataFrame({
        'id': [309],
        'answer': ['@fare_kurtosis[33.20] @age_values_within_one_stdev[516] @fare_skewness[4.78] @fare_values_within_one_stdev[818] @age_skewness[0.39] @age_kurtosis[0.17]']
    })
    answer_file = private / "answer.csv"
    answer.to_csv(answer_file, index=False)
    print(f"Created answer file: {answer_file}")

    # Verify
    assert train_file.exists(), "Training file not created"
    assert sample_submission_file.exists(), "Sample submission not created"
    assert answer_file.exists(), "Answer file not created"

    print(f"✓ DABench task 309 prepared successfully")



from pandas import DataFrame
from sklearn.metrics import root_mean_squared_error

from mlebench.grade_helpers import InvalidSubmissionError


def prepare_for_metric(submission: DataFrame, answers: DataFrame) -> dict:

    assert "fare_amount" in answers.columns, "Answers should have a fare_amount column"
    assert "key" in answers.columns, "Answers should have a key column"
    if "fare_amount" not in submission.columns:
        raise InvalidSubmissionError("Submission should have a fare_amount column")
    if "key" not in submission.columns:
        raise InvalidSubmissionError("Submission should have a key column")

    # Sort by 'key' to ensure alignment
    submission = submission.sort_values("key")
    answers = answers.sort_values("key")

    y_true = answers["fare_amount"]
    y_pred = submission["fare_amount"]

    return y_true, y_pred


def grade(submission: DataFrame, answers: DataFrame) -> float:
    y_true, y_pred = prepare_for_metric(submission, answers)
    return root_mean_squared_error(y_true, y_pred)



import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error

from mlebench.grade_helpers import InvalidSubmissionError


TARGET_COLUMNS = ['cohesion', 'syntax', 'vocabulary', 'phraseology', 'grammar', 'conventions']


def prepare_for_metric(submission: pd.DataFrame, answers: pd.DataFrame):
    """Prepare submission and answers for MCRMSE calculation."""
    if len(submission) != len(answers):
        raise InvalidSubmissionError(
            f"Submission length ({len(submission)}) != answers length ({len(answers)})"
        )

    if "text_id" not in submission.columns:
        raise InvalidSubmissionError("Submission must have a 'text_id' column")

    for col in TARGET_COLUMNS:
        if col not in submission.columns:
            raise InvalidSubmissionError(f"Submission must have a '{col}' column")

    # Sort by text_id
    submission = submission.sort_values("text_id").reset_index(drop=True)
    answers = answers.sort_values("text_id").reset_index(drop=True)

    # Check text_id alignment
    if not (submission["text_id"] == answers["text_id"]).all():
        raise InvalidSubmissionError("Submission and answers text_id mismatch")

    # Validate predictions
    for col in TARGET_COLUMNS:
        try:
            submission[col] = submission[col].astype(float)
        except (ValueError, TypeError):
            raise InvalidSubmissionError(f"'{col}' predictions must be numeric")

        if submission[col].isnull().any():
            raise InvalidSubmissionError(f"'{col}' predictions cannot contain NaN values")

    return submission, answers


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    """
    Calculate MCRMSE (Mean Columnwise Root Mean Squared Error).

    MCRMSE = mean of RMSE across all target columns
    """
    submission, answers = prepare_for_metric(submission, answers)

    rmse_scores = []
    for col in TARGET_COLUMNS:
        rmse = np.sqrt(mean_squared_error(answers[col], submission[col]))
        rmse_scores.append(rmse)

    mcrmse = np.mean(rmse_scores)
    return mcrmse



CLASSES = [
    "Acer_Capillipes",
    "Acer_Circinatum",
    "Acer_Mono",
    "Acer_Opalus",
    "Acer_Palmatum",
    "Acer_Pictum",
    "Acer_Platanoids",
    "Acer_Rubrum",
    "Acer_Rufinerve",
    "Acer_Saccharinum",
    "Alnus_Cordata",
    "Alnus_Maximowiczii",
    "Alnus_Rubra",
    "Alnus_Sieboldiana",
    "Alnus_Viridis",
    "Arundinaria_Simonii",
    "Betula_Austrosinensis",
    "Betula_Pendula",
    "Callicarpa_Bodinieri",
    "Castanea_Sativa",
    "Celtis_Koraiensis",
    "Cercis_Siliquastrum",
    "Cornus_Chinensis",
    "Cornus_Controversa",
    "Cornus_Macrophylla",
    "Cotinus_Coggygria",
    "Crataegus_Monogyna",
    "Cytisus_Battandieri",
    "Eucalyptus_Glaucescens",
    "Eucalyptus_Neglecta",
    "Eucalyptus_Urnigera",
    "Fagus_Sylvatica",
    "Ginkgo_Biloba",
    "Ilex_Aquifolium",
    "Ilex_Cornuta",
    "Liquidambar_Styraciflua",
    "Liriodendron_Tulipifera",
    "Lithocarpus_Cleistocarpus",
    "Lithocarpus_Edulis",
    "Magnolia_Heptapeta",
    "Magnolia_Salicifolia",
    "Morus_Nigra",
    "Olea_Europaea",
    "Phildelphus",
    "Populus_Adenopoda",
    "Populus_Grandidentata",
    "Populus_Nigra",
    "Prunus_Avium",
    "Prunus_X_Shmittii",
    "Pterocarya_Stenoptera",
    "Quercus_Afares",
    "Quercus_Agrifolia",
    "Quercus_Alnifolia",
    "Quercus_Brantii",
    "Quercus_Canariensis",
    "Quercus_Castaneifolia",
    "Quercus_Cerris",
    "Quercus_Chrysolepis",
    "Quercus_Coccifera",
    "Quercus_Coccinea",
    "Quercus_Crassifolia",
    "Quercus_Crassipes",
    "Quercus_Dolicholepis",
    "Quercus_Ellipsoidalis",
    "Quercus_Greggii",
    "Quercus_Hartwissiana",
    "Quercus_Ilex",
    "Quercus_Imbricaria",
    "Quercus_Infectoria_sub",
    "Quercus_Kewensis",
    "Quercus_Nigra",
    "Quercus_Palustris",
    "Quercus_Phellos",
    "Quercus_Phillyraeoides",
    "Quercus_Pontica",
    "Quercus_Pubescens",
    "Quercus_Pyrenaica",
    "Quercus_Rhysophylla",
    "Quercus_Rubra",
    "Quercus_Semecarpifolia",
    "Quercus_Shumardii",
    "Quercus_Suber",
    "Quercus_Texana",
    "Quercus_Trojana",
    "Quercus_Variabilis",
    "Quercus_Vulcanica",
    "Quercus_x_Hispanica",
    "Quercus_x_Turneri",
    "Rhododendron_x_Russellianum",
    "Salix_Fragilis",
    "Salix_Intergra",
    "Sorbus_Aria",
    "Tilia_Oliveri",
    "Tilia_Platyphyllos",
    "Tilia_Tomentosa",
    "Ulmus_Bergmanniana",
    "Viburnum_Tinus",
    "Viburnum_x_Rhytidophylloides",
    "Zelkova_Serrata",
]



import numpy as np
import pandas as pd
from sklearn.metrics import log_loss

from mlebench.competitions.utils import one_hot_dfs_to_log_loss_inputs
from mlebench.grade_helpers import InvalidSubmissionError

from .classes import CLASSES

tolerance = 1e-6


def prepare_for_metric(submission: pd.DataFrame, answers: pd.DataFrame) -> dict:
    id_col = "id"

    if submission.shape != (len(answers), len(CLASSES) + 1):
        raise InvalidSubmissionError(
            f"Submission shape {submission.shape} does not match answers shape {answers.shape}."
        )

    # assert that each row in submission sums to one, accounting for floating point errors
    if not np.all(np.isclose(submission.iloc[:, 1:].sum(axis=1), 1, atol=tolerance)):
        raise InvalidSubmissionError("Each row in submission should sum to one, as probabilities.")
    if not ((submission.iloc[:, 1:] >= 0) & (submission.iloc[:, 1:] <= 1)).all().all():
        raise InvalidSubmissionError(
            "All probabilities in submission DataFrame must be between 0 and 1."
        )

    log_loss_inputs = one_hot_dfs_to_log_loss_inputs(
        submission, answers, id_column=id_col, apply_softmax=False
    )

    return log_loss_inputs


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    log_loss_inputs = prepare_for_metric(submission, answers)
    return log_loss(**log_loss_inputs)



import pandas as pd
from scipy.sparse import csr_matrix
from sklearn.metrics import f1_score
from sklearn.preprocessing import MultiLabelBinarizer

from mlebench.grade_helpers import InvalidSubmissionError


def prepare_for_metric(
    submission: pd.DataFrame, answers: pd.DataFrame
) -> tuple[csr_matrix, csr_matrix]:
    """Transforms the submission and answers DataFrames into the required format for grading as sparse matrices."""

    # Answers checks
    assert (
        "id" in answers.columns and "attribute_ids" in answers.columns
    ), "Answers DataFrame must have 'id' and 'attribute_ids' columns"

    # Submission checks
    if "id" not in submission.columns:
        raise InvalidSubmissionError("Submission DataFrame must have an 'id' column.")
    if "attribute_ids" not in submission.columns:
        raise InvalidSubmissionError("Submission DataFrame must have a 'attribute_ids' column.")
    if len(submission) != len(answers):
        raise InvalidSubmissionError(
            "Submission and answers DataFrames must have the same number of rows."
        )

    # Match order
    submission = submission.sort_values(by="id").reset_index(drop=True)
    answers = answers.sort_values(by="id").reset_index(drop=True)
    if (submission["id"].values != answers["id"].values).any():
        raise InvalidSubmissionError(
            "Submission and answers DataFrames must have matching 'id' columns."
        )

    # pandas reads empty cells as NaNs, which are float. We fill with empty string to match type
    submission["attribute_ids"] = submission["attribute_ids"].fillna("")
    answers["attribute_ids"] = answers["attribute_ids"].fillna("")

    # Get classes
    classes = set(answers["attribute_ids"].str.split().explode().unique())

    # Convert to sparse matrices using MultiLabelBinarizer
    mlb = MultiLabelBinarizer(classes=sorted(classes), sparse_output=True)
    y_true = mlb.fit_transform(answers["attribute_ids"].str.split())
    y_pred = mlb.transform(submission["attribute_ids"].str.split())

    return y_true, y_pred


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    y_true, y_pred = prepare_for_metric(submission, answers)
    return f1_score(y_true=y_true, y_pred=y_pred, average="micro")



import pandas as pd
from pandas import DataFrame
from scipy.sparse import csr_matrix
from sklearn.metrics import f1_score
from sklearn.preprocessing import MultiLabelBinarizer

from mlebench.grade_helpers import InvalidSubmissionError
from mlebench.utils import get_logger

logger = get_logger(__name__)


def grade(submission: DataFrame, answers: DataFrame) -> float:
    """Grades the submission against the test set."""

    y_true, y_pred = prepare_for_metric(submission, answers)
    return f1_score(y_true=y_true, y_pred=y_pred, average="micro")


def prepare_for_metric(
    submission: pd.DataFrame, answers: pd.DataFrame
) -> tuple[csr_matrix, csr_matrix]:

    # Answer checks
    assert "Id" in answers.columns, "Answers must have an 'Id' column."
    assert "Tags" in answers.columns, "Answers must have a 'Tags' column."

    # Submission checks
    if "Id" not in submission.columns:
        raise InvalidSubmissionError("Submission DataFrame must have an 'Id' column.")
    if "Tags" not in submission.columns:
        raise InvalidSubmissionError("Submission DataFrame must have a 'Tags' column.")
    if len(submission) != len(answers):
        raise InvalidSubmissionError(
            "Submission DataFrame must have the same number of rows as the answers DataFrame."
        )

    # Match order
    submission = submission.sort_values(by="Id").reset_index(drop=True)
    answers = answers.sort_values(by="Id").reset_index(drop=True)
    if (submission["Id"].values != answers["Id"].values).any():
        raise InvalidSubmissionError("Submission and answers must have matching 'Id's.")

    # Get classes
    classes = set()

    for tags in answers["Tags"]:
        if not isinstance(tags, str):
            logger.warning(f"Tags from answers '{tags}' not of type str! Skipping.")
            continue

        tags_split = tags.split()
        classes.update(tags_split)

    # Convert to sparse matrix using MultiLabelBinarizer
    mlb = MultiLabelBinarizer(classes=sorted(classes), sparse_output=True)
    y_true = mlb.fit_transform(answers["Tags"].fillna("").str.split())
    y_pred = mlb.transform(submission["Tags"].fillna("").str.split())

    return y_true, y_pred



from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def _create_and_save_split(
    df_to_split: pd.DataFrame,
    public_path: Path,
    private_path: Path,
    test_size: int,
    random_state: int,
):
    """
    Helper function to perform a data split, re-index, and save all required files.

    Args:
        df_to_split (pd.DataFrame): The dataframe to be split.
        public_path (Path): The directory for public-facing files (train, test features).
        private_path (Path): The directory for private-facing files (test labels).
        test_size (int): The number of samples for the test set.
        random_state (int): The random state for reproducibility.

    Returns:
        pd.DataFrame: The new, smaller training dataframe from the split.
    """
    # Create output directories if they don't exist
    public_path.mkdir(parents=True, exist_ok=True)
    private_path.mkdir(parents=True, exist_ok=True)

    # Perform the split
    train_df, test_df = train_test_split(
        df_to_split, test_size=test_size, random_state=random_state
    )

    # make ids go from 0 to len(train_df) - 1
    train_df.id = np.arange(len(train_df))
    # and from len(train_df) to len(train_df) + len(test_df) - 1
    test_df.id = np.arange(len(train_df), len(train_df) + len(test_df))

    # make downstream files
    test_df_without_labels = test_df.drop(columns=["target"]).copy()
    gold_submission = test_df[["id", "target"]].copy()
    sample_submission = gold_submission.copy()
    sample_submission.target = 0.5

    # save files with required names
    train_df.to_csv(public_path / "train.csv", index=False)
    test_df.to_csv(private_path / "test.csv", index=False)
    test_df_without_labels.to_csv(public_path / "test.csv", index=False)
    gold_submission.to_csv(private_path / "gold_submission.csv", index=False)
    sample_submission.to_csv(public_path / "sample_submission.csv", index=False)

    # run checks for this split
    assert len(train_df) + len(test_df) == len(
        df_to_split
    ), "Expected the sum of the lengths of the new train and test to be equal to the length of the input data."
    assert len(test_df) == len(
        sample_submission
    ), "Expected the length of the private test to be equal to the length of the sample submission."
    assert len(test_df) == len(
        gold_submission
    ), "Expected the length of the private test to be equal to the length of the gold submission."
    assert (
        train_df.columns.to_list() == df_to_split.columns.to_list()
    ), "Expected the columns of the new train to be the same as the columns of the input data."
    assert (
        test_df.columns.to_list() == df_to_split.columns.to_list()
    ), "Expected the columns of the new test to be the same as the columns of the input data"
    assert set(train_df.id).isdisjoint(
        set(test_df.id)
    ), "Expected the ids of the new train and test to be disjoint."

    return train_df


def prepare(raw: Path, public: Path, private: Path):

    old_train = read_csv(raw / "train.csv")

    # --- Stage 1: Create the original train/test split ---
    # This section produces the primary competition data. Its outputs in `public/`
    # and `private/` are guaranteed to be identical to the original script.

    # 900k train, 1.6m - 900k = 700k test; so 700k/1.6m = 0.4375
    # We create our split at 100,000 test samples to get same OOM while keeping as many samples as possible in train
    new_train = _create_and_save_split(
        df_to_split=old_train,
        public_path=public,
        private_path=private,
        test_size=100_000,
        random_state=0,
    )

    # --- Stage 2: Create the new train/validation split ---
    # This section takes the `new_train` set from Stage 1 and splits it again
    # to create a smaller training set and a new validation set.
    # The outputs are saved to parallel `public_val/` and `private_val/` directories.

    # Define paths for the new validation set directories, parallel to the original ones.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Split `new_train` again using the exact same logic and test size.
    # The resulting `test_val` set will have 100,000 samples, same as the original `test` set.
    # The filenames and directory structure are mirrored for consistency.
    _create_and_save_split(
        df_to_split=new_train,
        public_path=public_val,
        private_path=private_val,
        test_size=100_000,
        random_state=0,
    )


import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm


def make_image_subpath(image_id: str) -> Path:
    """
    Creates a triple-nested directory structure from the first 3 characters of the image_id.
    """
    subpath = Path(image_id[0]) / image_id[1] / image_id[2] / f"{image_id}.png"
    return subpath


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    """
    # Load train data
    old_train = pd.read_csv(raw / "train_labels.csv")

    # Create train, test from train split
    new_train, new_test = train_test_split(old_train, test_size=0.2, random_state=0)
    new_train.to_csv(public / "train_labels.csv", index=False)
    new_test.to_csv(private / "test.csv", index=False)

    # Copy train files
    for idx, row in tqdm(new_train.iterrows(), total=len(new_train), desc="Copying train images"):
        image_id = row["image_id"]
        src = raw / "train" / make_image_subpath(image_id)
        dst = public / "train" / make_image_subpath(image_id)
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(src=src, dst=dst)

    # Copy test files
    for idx, row in tqdm(new_test.iterrows(), total=len(new_test), desc="Copying test images"):
        image_id = row["image_id"]
        src = raw / "train" / make_image_subpath(image_id)
        dst = public / "test" / make_image_subpath(image_id)
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(src=src, dst=dst)

    # Create sample submission
    sample_submission = new_test.copy()
    sample_submission["InChI"] = "InChI=1S/H2O/h1H2"
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    # Copy other files in the dataset (no modification needed)
    shutil.copyfile(src=raw / "extra_approved_InChIs.csv", dst=public / "extra_approved_InChIs.csv")

    # Checks
    assert len(new_train) + len(new_test) == len(
        old_train
    ), f"Expected {len(old_train)} total images in new_train ({len(new_train)}) and new_test ({len(new_test)})"
    assert len(list((public / "train").glob("**/*.png"))) == len(
        new_train
    ), f"Expected {len(new_train)} train images in public/train, but got {len(list((public / 'train').glob('**/*.png')))}"
    assert len(list((public / "test").glob("**/*.png"))) == len(
        new_test
    ), f"Expected {len(new_test)} test images in public/test, but got {len(list((public / 'test').glob('**/*.png')))}"

    assert "image_id" in sample_submission.columns, "Sample submission must have 'image_id' column"
    assert "InChI" in sample_submission.columns, "Sample submission must have 'InChI' column"
    assert len(sample_submission) == len(
        new_test
    ), f"Expected {len(new_test)} images in sample submission, but got {len(sample_submission)}"



import shutil
from pathlib import Path

from sklearn.model_selection import train_test_split
from tqdm.auto import tqdm

from mlebench.competitions.utils import get_ids_from_tf_records
from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path):

    # need to split based on the TFRecord files, since not mentioned in the CSVs
    tfrecord_files = [
        path
        for path in sorted((raw / "train_tfrecords").iterdir())
        if path.is_file() and path.suffix == ".tfrec"
    ]

    # In the original there are 21397 train samples and they say test has ~15000 test samples, which is ~ 0.4/0.6 test/train split
    # We use 0.1 ratio to avoid removing too many samples from train
    new_train_tfrecords, new_test_tfrecords = train_test_split(
        tfrecord_files, test_size=0.1, random_state=0
    )

    # parse the IDs from the test tf records
    test_ids = []
    for path in new_test_tfrecords:
        test_ids.extend(get_ids_from_tf_records(path))

    old_train = read_csv(raw / "train.csv")

    old_train["split"] = "train"
    old_train.loc[old_train["image_id"].isin(test_ids), "split"] = "test"

    new_train = old_train[old_train["split"] == "train"].drop(columns=["split"])
    new_test = old_train[old_train["split"] == "test"].drop(columns=["split"])

    sample_submission = new_test.copy()
    sample_submission["label"] = 4

    new_train.to_csv(public / "train.csv", index=False)
    new_test.to_csv(private / "test.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    (public / "train_tfrecords").mkdir(parents=True, exist_ok=True)
    for i, path in tqdm(
        enumerate(new_train_tfrecords),
        desc="Copying Train TFRecords",
        total=len(new_train_tfrecords),
    ):
        length = path.stem.split("-")[1]
        new_name = f"ld_train{i:02d}-{length}.tfrec"

        shutil.copy(path, public / "train_tfrecords" / new_name)

    (public / "test_tfrecords").mkdir(parents=True, exist_ok=True)
    for i, path in tqdm(
        enumerate(new_test_tfrecords), desc="Copying Test TFRecords", total=len(new_test_tfrecords)
    ):
        length = path.stem.split("-")[1]
        new_name = f"ld_test{i:02d}-{length}.tfrec"

        shutil.copy(path, public / "test_tfrecords" / new_name)

    (public / "train_images").mkdir(parents=True, exist_ok=True)
    for image_id in tqdm(new_train["image_id"], desc="Copying Train Images", total=len(new_train)):
        shutil.copy(raw / "train_images" / image_id, public / "train_images")

    (public / "test_images").mkdir(parents=True, exist_ok=True)
    for image_id in tqdm(new_test["image_id"], desc="Copying Test Images", total=len(new_test)):
        shutil.copy(raw / "train_images" / image_id, public / "test_images")

    shutil.copy(raw / "label_num_to_disease_map.json", public / "label_num_to_disease_map.json")

    # checks
    assert len(new_train) + len(new_test) == len(
        old_train
    ), "Expected new train and new test lengths to sum to old train length"
    assert len(sample_submission) == len(
        new_test
    ), "Expected sample submission length to be equal to new test length"

    assert len(new_train) == sum(
        1 for _ in (public / "train_images").iterdir()
    ), "Mismatch in number of expected train images copied"
    assert len(new_test) == sum(
        1 for _ in (public / "test_images").iterdir()
    ), "Mismatch in number of expected test images copied"

    assert len(new_train_tfrecords) == sum(
        1 for _ in (public / "train_tfrecords").iterdir()
    ), "Mismatch in number of expected train TFRecords copied"
    assert len(new_test_tfrecords) == sum(
        1 for _ in (public / "test_tfrecords").iterdir()
    ), "Mismatch in number of expected test TFRecords copied"

    assert new_train.columns.tolist() == [
        "image_id",
        "label",
    ], "Expected new train columns to be ['image_id', 'label']"
    assert new_test.columns.tolist() == [
        "image_id",
        "label",
    ], "Expected new test columns to be ['image_id', 'label']"
    assert sample_submission.columns.tolist() == [
        "image_id",
        "label",
    ], "Expected sample submission columns to be ['image_id', 'label']"

    assert set(new_train["image_id"]).isdisjoint(
        new_test["image_id"]
    ), "Expected train and test image IDs to be disjoint"



import csv
import zipfile
from pathlib import Path

from sklearn.model_selection import train_test_split

from mlebench.utils import compress, extract, read_csv


def prepare(raw: Path, public: Path, private: Path):

    # Extract
    extract(raw / "ru_test_2.csv.zip", raw)  # We only use the 2nd stage test set
    extract(raw / "ru_train.csv.zip", raw)
    extract(raw / "ru_sample_submission_2.csv.zip", raw)

    # Create train and test splits from train set
    old_train = read_csv(raw / "ru_train.csv")
    # We split so that we don't share any sentence_ids between train and test
    # This gives us len(new_train) = 9515325 and len(answers) = 1059191
    unique_sentence_ids = old_train["sentence_id"].unique()
    train_sentence_ids, test_sentence_ids = train_test_split(
        unique_sentence_ids, test_size=0.1, random_state=0
    )
    new_train = old_train[old_train["sentence_id"].isin(train_sentence_ids)]
    answers = old_train[old_train["sentence_id"].isin(test_sentence_ids)]
    assert set(new_train["sentence_id"]).isdisjoint(
        set(answers["sentence_id"])
    ), f"sentence_id is not disjoint between train and test sets"

    # "sentence_id" counts need to be reset for new_train and answers
    new_train_id_mapping = {
        old_id: new_id for new_id, old_id in enumerate(new_train["sentence_id"].unique())
    }
    new_train["sentence_id"] = new_train["sentence_id"].map(new_train_id_mapping)
    answers_id_mapping = {
        old_id: new_id for new_id, old_id in enumerate(answers["sentence_id"].unique())
    }
    answers["sentence_id"] = answers["sentence_id"].map(answers_id_mapping)

    # Create new test set
    new_test = answers.drop(["after", "class"], axis=1).copy()

    # Reformat answers to match sample submission format
    answers = answers[["sentence_id", "token_id", "after"]].copy()
    answers["id"] = answers["sentence_id"].astype(str) + "_" + answers["token_id"].astype(str)
    answers = answers[["id", "after"]]

    # Create sample submission
    sample_submission = new_test[["sentence_id", "token_id", "before"]].copy()
    sample_submission["id"] = (
        sample_submission["sentence_id"].astype(str)
        + "_"
        + sample_submission["token_id"].astype(str)
    )
    sample_submission["after"] = sample_submission["before"]
    sample_submission = sample_submission[["id", "after"]]

    # Checks
    assert new_train.columns.tolist() == [
        "sentence_id",
        "token_id",
        "class",
        "before",
        "after",
    ], f"new_train.columns.tolist() == {new_train.columns.tolist()}"
    assert new_test.columns.tolist() == [
        "sentence_id",
        "token_id",
        "before",
    ], f"new_test.columns.tolist() == {new_test.columns.tolist()}"
    assert sample_submission.columns.tolist() == [
        "id",
        "after",
    ], f"sample_submission.columns.tolist() == {sample_submission.columns.tolist()}"
    assert answers.columns.tolist() == [
        "id",
        "after",
    ], f"answers.columns.tolist() == {answers.columns.tolist()}"
    assert len(new_test) + len(new_train) == len(
        old_train
    ), f"New train and test sets do not sum to old train set, got {len(new_test) + len(new_train)} and {len(old_train)}"

    # Write CSVs
    answers.to_csv(
        private / "answers.csv", index=False, quotechar='"', quoting=csv.QUOTE_NONNUMERIC
    )
    sample_submission.to_csv(
        private / "sample_submission.csv", index=False, quotechar='"', quoting=csv.QUOTE_NONNUMERIC
    )
    new_train.to_csv(
        public / "ru_train.csv", index=False, quotechar='"', quoting=csv.QUOTE_NONNUMERIC
    )
    new_test.to_csv(
        public / "ru_test_2.csv", index=False, quotechar='"', quoting=csv.QUOTE_NONNUMERIC
    )
    sample_submission.to_csv(
        public / "ru_sample_submission_2.csv",
        index=False,
        quotechar='"',
        quoting=csv.QUOTE_NONNUMERIC,
    )

    # Zip up
    with zipfile.ZipFile(public / "ru_train.csv.zip", "w") as zipf:
        zipf.write(public / "ru_train.csv", arcname="ru_train.csv")
    with zipfile.ZipFile(public / "ru_test_2.csv.zip", "w") as zipf:
        zipf.write(public / "ru_test_2.csv", arcname="ru_test_2.csv")
    with zipfile.ZipFile(public / "ru_sample_submission_2.csv.zip", "w") as zipf:
        zipf.write(public / "ru_sample_submission_2.csv", arcname="ru_sample_submission_2.csv")
    (public / "ru_train.csv").unlink()
    (public / "ru_test_2.csv").unlink()
    (public / "ru_sample_submission_2.csv").unlink()



# dsat/services/data_analyzer.py

import logging
import traceback
from pathlib import Path
import pandas as pd
from typing import Optional, List
from itertools import islice

from dsat.models.task import TaskType

logger = logging.getLogger(__name__)


def generate_file_tree(
    start_path: Path,
    max_depth: int = 5,
    max_files: int = 100,
    max_items_per_dir: int = 20,
    display_root_name: Optional[str] = None
) -> str:
    """
    Generates a textual representation of the file tree with intelligent truncation.

    This version prevents a single large directory from consuming the entire file limit,
    and it filters out common noise files.

    Args:
        start_path: The root directory to start the tree from.
        max_depth: The maximum depth to traverse into directories.
        max_files: The global limit for the total number of files to display.
        max_items_per_dir: The maximum number of items (files and dirs) to show per directory.
        display_root_name: An optional name to display for the root directory.
    """
    tree = []
    start_path = Path(start_path)
    if not start_path.exists():
        return f"Directory not found: {start_path}"


    base_name = display_root_name if display_root_name is not None else start_path.name
    file_count = 0
    global_limit_reached = False

    def _walk(path: Path, prefix: str, depth: int):
        nonlocal file_count, global_limit_reached
        
        if depth > max_depth or global_limit_reached:
            return

        try:
            # Avoid listing huge directories fully (e.g., image folders).
            # We only sample a small number of entries for display.
            sampled = list(islice(path.iterdir(), max_items_per_dir + 1))
        except OSError as e:
            logger.warning(f"Error reading directory {path}: {e}")
            tree.append(f"{prefix}└── [Error reading directory]")
            return

        truncated_in_dir = len(sampled) > max_items_per_dir
        if truncated_in_dir:
            display_items = sampled[: max(1, max_items_per_dir // 2)]
        else:
            display_items = sampled
        display_items = sorted(display_items, key=lambda p: p.name)

        pointers = ['├── '] * (len(display_items) - 1) + ['└── ']
        # If we truncated this directory, the last visible item is not the true last item
        if truncated_in_dir:
            pointers[-1] = '├── '

        for pointer, sub_path in zip(pointers, display_items):
            if global_limit_reached:
                return

            if not sub_path.is_dir():
                # Check global file limit *before* adding the next file
                if file_count >= max_files:
                    global_limit_reached = True
                    return
                file_count += 1

            display_name = sub_path.name + ('/' if sub_path.is_dir() else '')
            tree.append(f"{prefix}{pointer}{display_name}")

            if sub_path.is_dir():
                extension = '│   ' if pointer == '├── ' else '    '
                _walk(sub_path, prefix=prefix + extension, depth=depth + 1)

        if truncated_in_dir:
            tree.append(f"{prefix}└── [... more items truncated ...]")

    tree.append(f"{base_name}/")
    _walk(start_path, prefix="", depth=1)

    if global_limit_reached:
        tree.append(f"\n[... Truncated. Total file limit ({max_files}) reached ...]")

    return "\n".join(tree)


class DataAnalyzer:
    """
    A centralized service for analyzing input data directories and generating
    a comprehensive textual overview for the Agent.
    """
    def analyze(self, data_dir: Path, output_filename: str, task_type: Optional[TaskType] = None, optimization_context: bool = False) -> str:
        """
        Analyzes the data directory and returns a formatted overview string,
        including critical I/O instructions.
        """
        report = self.analyze_data(data_dir, task_type)
        report += self.generate_io_instructions(output_filename, optimization_context)
        return report

    def analyze_data(self, data_dir: Path, task_type: Optional[TaskType] = None) -> str:
        """
        Analyzes the data directory and returns ONLY the data report (structure, format, etc.).
        Does NOT include I/O instructions.
        """
        if not data_dir or not data_dir.exists() or not data_dir.is_dir():
            logger.error(f"Data directory issue during analysis: {data_dir}")
            return "Error: Input data directory not found, not provided, or is not a directory."

        report = "\n\n--- COMPREHENSIVE DATA REPORT ---\n\n"

        # 1. Analyze File Structure (Universal)
        report += self._analyze_structure(data_dir)

        report += self._analyze_data_schema(data_dir)

        # 3. Task-Specific Analysis
        if task_type == "kaggle":
            submission_analysis = self._analyze_kaggle_submission_format(data_dir)
            if submission_analysis:
                report += f"## Submission Format Requirements\n{submission_analysis}\n\n"
        
        # Remove the call to _generate_io_instructions from here
        return report

    def generate_io_instructions(self, output_filename: str, optimization_context: bool = False) -> str:
        """
        Generate standardized I/O instructions reflecting that CWD is the input directory.
        """
        output_suffix = Path(output_filename).suffix.lower()

        input_instructions = (
            "1. **INPUT DATA:**\n"
            "   - All input files are located in the **current working directory** (./).\n"
            "   - Example: Use `pd.read_csv('train.csv')`."
        )

        if optimization_context:
            example_write = ""
            if output_suffix == ".csv":
                example_write = "   - **Example Write (Conceptual):** `final_df.to_csv(output_path.name, index=False)`"
            elif output_suffix == ".npy":
                example_write = "   - **Example Write (Conceptual):** `np.save(output_path.name, preds)`"

            output_instructions = (
                "2. **OUTPUT FILE (Dynamic Workflow Context):**\n"
                "   - Your workflow's `solve` method receives an `output_path` argument.\n"
                "   - You MUST save your final submission file using the filename derived from this argument (e.g., `output_path.name`).\n"
                "   - The file must be saved in the current working directory (./).\n"
                + (f"\n{example_write}" if example_write else "")
            )
        else:
            example_write = ""
            if output_suffix == ".csv":
                example_write = f"   - **Correct Example:** `submission_df.to_csv('{output_filename}', index=False)`"
            elif output_suffix == ".npy":
                example_write = f"   - **Correct Example:** `np.save('{output_filename}', preds)`"

            output_instructions = (
                f"2. **OUTPUT FILE:**\n"
                f"   - You MUST save your final submission file to the **current working directory** (./).\n"
                f"   - The required output filename is: `{output_filename}`\n"
                + (f"{example_write}\n" if example_write else "")
            )

        return f"""
--- CRITICAL I/O REQUIREMENTS ---

You MUST follow these file system rules precisely. Failure to do so will cause a fatal error.

{input_instructions}

{output_instructions}

**IMPORTANT:** These path requirements are non-negotiable and must be followed exactly.
"""

    def _analyze_structure(self, data_dir: Path) -> str:
        """Generates the file tree representation."""
        try:
            tree_output = generate_file_tree(data_dir, display_root_name=".")
            return f"## Directory Structure (Current Working Directory)\n```text\n{tree_output}\n```\n\n"
        except Exception as e:
            logger.error(f"Failed to generate file tree for {data_dir}: {traceback.format_exc()}")
            return f"## Directory Structure\nError analyzing structure: {traceback.format_exc()}\n\n"

    def _analyze_kaggle_submission_format(self, data_dir: Path) -> str:
        """
        Analyzes the sample submission file for Kaggle tasks and extracts a
        prescriptive schema.
        """
        sample_submission_file = self._find_sample_submission(data_dir)

        if not sample_submission_file:
            return ""

        try:
            # Read the sample submission file to get format details
            sample_df = pd.read_csv(sample_submission_file)

            # Get the first few rows and data types
            head_info = sample_df.head().to_string(index=False)
            dtypes_info = sample_df.dtypes.to_string()
            
            required_columns = sample_df.columns.tolist()
            columns_instruction = f"""
**Required Submission Columns:**
Your submission file MUST contain the following columns in this exact order:
```
{required_columns}
```
This is a strict requirement for the submission to be graded correctly. The grading system uses the non-prediction columns (like 'Comment' or an 'id') to match your predictions against the ground truth.
"""

            return f"""
**CRITICAL:** Your final submission file MUST EXACTLY match the format of the sample submission file provided (`{sample_submission_file.name}`).
This includes the column names, column order, and data types. Failure to adhere to this format will result in a score of zero.

{columns_instruction}

**Format Details:**
*First 5 rows:*
```text
{head_info}
```

*Data types:*
```text
{dtypes_info}
```

"""
        except Exception as e:
            # Fallback if pandas fails
            logger.warning(f"Could not read sample submission file '{sample_submission_file}' for detailed analysis: {traceback.format_exc()}")
            return f"""
**CRITICAL:** Your final submission file MUST match the format of the sample submission file: `{sample_submission_file.name}`.
(Note: Automatic format analysis failed, please inspect the file manually).
"""

    def _find_sample_submission(self, data_dir: Path) -> Optional[Path]:
        """Helper to locate the sample submission file."""
        try:
            for file in data_dir.iterdir():
                file_name_lower = file.name.lower()
                if "sample" in file_name_lower and "submission" in file_name_lower and file_name_lower.endswith(".csv"):
                    return file
        except Exception as e:
            logger.warning(f"Could not scan data directory '{data_dir}': {traceback.format_exc()}")
        return None

    def _analyze_data_schema(self, data_dir: Path) -> str:
        """
        Analyzes the schema of potential training and testing files to provide a
        structured overview of columns, data types, missing values, and cardinality.
        This helps the agent make better decisions about preprocessing.
        """
        report_parts = []
        # Define supported extensions and keywords for more robust discovery
        SUPPORTED_EXTENSIONS = ('.csv', '.tsv', '.parquet', '.xlsx')
        KEYWORDS = ('train', 'test', 'val', 'eval', 'sample', 'submission', 'sub', 'data')

        # Fast path: most prepared competitions keep train/test-like tables at the root.
        files_to_analyze: List[Path] = []
        try:
            root_files = [p for p in data_dir.iterdir() if p.is_file()]
        except OSError:
            root_files = []

        for p in root_files:
            if p.suffix.lower() not in SUPPORTED_EXTENSIONS:
                continue
            if any(keyword in p.stem.lower() for keyword in KEYWORDS):
                files_to_analyze.append(p)

        # Fallback: if still nothing, just take the first few supported files at root
        if not files_to_analyze and root_files:
            files_to_analyze = [p for p in root_files if p.suffix.lower() in SUPPORTED_EXTENSIONS][:3]

        # Deep discoveryFallback: bounded recursive search (avoid walking huge image folders).
        if not files_to_analyze:
            max_depth = 3
            max_dirs = 200
            max_files = 8
            per_dir_limit = 200

            queue: List[tuple[Path, int]] = [(data_dir, 0)]
            visited = 0
            while queue and visited < max_dirs and len(files_to_analyze) < max_files:
                current, depth = queue.pop(0)
                visited += 1
                if depth > max_depth:
                    continue
                try:
                    sampled = list(islice(current.iterdir(), per_dir_limit + 1))
                except OSError:
                    continue

                truncated = len(sampled) > per_dir_limit
                entries = sampled[:per_dir_limit] if truncated else sampled

                for entry in entries:
                    if entry.is_dir():
                        # If this directory is huge, avoid descending further.
                        if truncated and depth >= 1:
                            continue
                        queue.append((entry, depth + 1))
                        continue
                    if entry.suffix.lower() not in SUPPORTED_EXTENSIONS:
                        continue
                    if not any(keyword in entry.stem.lower() for keyword in KEYWORDS):
                        continue
                    files_to_analyze.append(entry)
                    if len(files_to_analyze) >= max_files:
                        break

        files_to_analyze = sorted(set(files_to_analyze))

        if not files_to_analyze:
            return ""

        max_rows = 5000
        for file_path in files_to_analyze:
            try:
                # Dynamically choose the reader based on file extension
                ext = file_path.suffix.lower()
                if ext in ['.csv', '.tsv']:
                    sep = "\t" if ext == ".tsv" else ","
                    # Try multiple encodings
                    df = None
                    for enc in ['utf-8', 'gbk', 'latin1', 'utf-8-sig']:
                        try:
                            df = pd.read_csv(file_path, sep=sep, nrows=max_rows, encoding=enc)
                            break
                        except (UnicodeDecodeError, Exception):
                            continue
                    
                    if df is None:
                        raise Exception(f"Failed to read CSV with multiple encodings for {file_path.name}")
                elif ext == '.parquet':
                    # Note: This requires 'pyarrow' or 'fastparquet' to be installed
                    df = pd.read_parquet(file_path).head(max_rows)
                else:
                    # Skip unsupported but matched files
                    continue

                report_parts.append(f"### Analysis of `{file_path.relative_to(data_dir)}`")

                # Create a summary DataFrame
                summary = pd.DataFrame({
                    'Data Type': df.dtypes,
                    'Missing (%)': (df.isnull().sum() * 100 / len(df)).round(2),
                    'Cardinality': df.nunique(),
                })
                
                sample_values = [col.dropna().head(2).tolist() for _, col in df.items()]
                summary['Sample Values'] = sample_values

                # Truncate sample values for readability
                summary['Sample Values'] = summary['Sample Values'].apply(
                    lambda x: str(x) if len(str(x)) < 40 else str(x)[:37] + '...'
                )

                report_parts.append(f"```text\n{summary.to_string()}\n```")
            except Exception as e:
                logger.warning(f"Could not analyze schema for {file_path.name}: {traceback.format_exc()}")
                report_parts.append(f"### Analysis of `{file_path.relative_to(data_dir)}`\nCould not be analyzed due to error: {e}")

        if not report_parts:
            return ""

        return "## Data Schema Analysis\n" + "\n\n".join(report_parts) + "\n\n"



from pathlib import Path

from sklearn.model_selection import train_test_split

from mlebench.competitions.utils import df_to_one_hot
from mlebench.utils import extract, read_csv

from .classes import CLASSES


def prepare(raw: Path, public: Path, private: Path):
    target_col = "author"
    id_col = "id"

    # extract only what we need
    extract(raw / "train.zip", raw)

    # =================================================================
    # == Original Data Split (for `public` and `private` directories)
    # =================================================================
    # Create main train/test split from the raw data
    old_train = read_csv(raw / "train.csv")
    train_main, test_main = train_test_split(old_train, test_size=0.1, random_state=0)
    test_main_without_labels = test_main.drop(columns=[target_col])

    # private test matches the format of sample submission
    one_hot_test_main = df_to_one_hot(
        test_main.drop(columns=["text"]),
        id_column=id_col,
        target_column=target_col,
        classes=CLASSES,
    )
    # fill the sample submission with arbitrary values (matching kaggle.com)
    sample_submission = one_hot_test_main.copy()
    sample_submission["EAP"] = 0.403493538995863
    sample_submission["HPL"] = 0.287808366106543
    sample_submission["MWS"] = 0.308698094897594

    # save files to original public/private directories
    train_main.to_csv(public / "train.csv", index=False)
    test_main_without_labels.to_csv(public / "test.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)
    one_hot_test_main.to_csv(private / "test.csv", index=False)

    # =====================================================================
    # == New Validation Split (for `public_val` and `private_val` dirs)
    # =====================================================================
    # Define and create the new parallel directories for the validation set
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"
    public_val.mkdir(parents=True, exist_ok=True)
    private_val.mkdir(parents=True, exist_ok=True)

    # Split the main training set again to create a new, smaller training set and a validation set.
    # The logic and random_state are replicated to ensure the validation set size is
    # consistent with the main test set size.
    train_val, test_val = train_test_split(train_main, test_size=0.1, random_state=0)
    test_val_without_labels = test_val.drop(columns=[target_col])

    # Process the new validation set's labels, mirroring the original logic
    one_hot_test_val = df_to_one_hot(
        test_val.drop(columns=["text"]),
        id_column=id_col,
        target_column=target_col,
        classes=CLASSES,
    )

    # Create a new sample submission for the validation set
    sample_submission_val = one_hot_test_val.copy()
    sample_submission_val["EAP"] = 0.403493538995863
    sample_submission_val["HPL"] = 0.287808366106543
    sample_submission_val["MWS"] = 0.308698094897594

    # Save files to the new validation directories using identical filenames
    train_val.to_csv(public_val / "train.csv", index=False)
    test_val_without_labels.to_csv(public_val / "test.csv", index=False)
    sample_submission_val.to_csv(public_val / "sample_submission.csv", index=False)
    one_hot_test_val.to_csv(private_val / "test.csv", index=False)


import json
import random
import shutil
import tarfile
from pathlib import Path
from typing import Dict, List

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import get_logger

logger = get_logger(__name__)


def add_to_tar(src: Path, out: Path):

    assert src.exists(), f"Source directory `{src}` does not exist."
    assert src.is_dir(), f"Expected a directory, but got `{src}`."

    tqdm_desc = f"Taring {src.name} to {out.name}"
    file_paths = [path for path in src.rglob("*") if path.is_file()]
    total_files = len(file_paths)

    with tarfile.open(out, "w") as tar:
        for file_path in tqdm(file_paths, desc=tqdm_desc, unit="file", total=total_files):
            tar.add(file_path, arcname=file_path.relative_to(src))


def _split_by_category(
    data_by_category: Dict, test_size: float, random_state: int
) -> tuple[Dict, Dict]:
    """Splits data for each category into train and test sets using the original script's logic."""
    train_split_by_category = {}
    test_split_by_category = {}

    for category_id, annotation_image_metadata in tqdm(
        data_by_category.items(),
        desc=f"Assigning train/test splits (test_size={test_size:.3f})",
    ):
        n_samples = len(annotation_image_metadata)
        if n_samples == 1:
            # If only one sample, put it in train
            train_annotations_images = annotation_image_metadata
            test_annotations_images = []
        elif n_samples < 5:  # Minimum 5 samples to ensure at least 1 in test
            num_test_samples = max(1, int(n_samples * test_size))
            train_annotations_images = annotation_image_metadata[:-num_test_samples]
            test_annotations_images = annotation_image_metadata[-num_test_samples:]
        else:
            train_annotations_images, test_annotations_images = train_test_split(
                annotation_image_metadata, test_size=test_size, random_state=random_state
            )

        train_split_by_category[category_id] = train_annotations_images
        test_split_by_category[category_id] = test_annotations_images

    return train_split_by_category, test_split_by_category


def _generate_split_files(
    train_annotation_image_metadata_by_category: Dict,
    test_annotation_image_metadata_by_category: Dict,
    old_train_metadata: Dict,
    raw_path: Path,
    public_path: Path,
    private_path: Path,
    dev_mode: bool,
    image_count: int,
):
    """
    Processes a given train/test split and saves all corresponding files
    (metadata, images, private answers, etc.) to the specified output directories.
    """
    public_path.mkdir(parents=True, exist_ok=True)
    private_path.mkdir(parents=True, exist_ok=True)

    # Create new train2019.json
    new_train_metadata = (
        old_train_metadata.copy()
    )  # Keep 'info', 'categories', 'licenses' unchanged
    new_train_metadata.update(
        {
            "annotations": [],
            "images": [],
        }
    )
    for category_id, annotation_image_metadata in tqdm(
        train_annotation_image_metadata_by_category.items(),
        desc=f"[{public_path.name}] Creating new train2019.json",
        total=len(train_annotation_image_metadata_by_category),
    ):
        for annotation_image in annotation_image_metadata:
            new_annotation = annotation_image["annotation"].copy()
            new_train_metadata["annotations"].append(new_annotation)
            new_image = annotation_image["image"].copy()
            new_train_metadata["images"].append(new_image)

    with open(public_path / "train2019.json", "w") as f:
        json.dump(new_train_metadata, f, indent=4, sort_keys=True)

    # Copy over val2019.json
    shutil.copy(raw_path / "val2019.json", public_path / "val2019.json")

    # Create new test2019.json
    new_to_old_file_name = {}
    new_test_metadata = old_train_metadata.copy()
    del new_test_metadata["categories"]
    new_test_metadata.update(
        {
            "annotations": [],
            "images": [],
        }
    )
    # Flatten and shuffle test set so that we don't have all the same categories in a row
    test_annotations_images = [
        item
        for sublist in test_annotation_image_metadata_by_category.values()
        for item in sublist
    ]
    random.Random(0).shuffle(test_annotations_images)
    for idx, annotation_image in tqdm(
        enumerate(test_annotations_images),
        desc=f"[{public_path.name}] Creating new test2019.json",
        total=len(test_annotations_images),
    ):
        new_annotation = annotation_image["annotation"].copy()
        new_test_metadata["annotations"].append(new_annotation)

        new_image = annotation_image["image"].copy()
        old_file_name = new_image["file_name"]
        # go from e.g. "train_val2019/Plants/400/d1322d13ccd856eb4236c8b888546c79.jpg" to "test2019/d1322d13ccd856eb4236c8b888546c79.jpg"
        new_file_name = "test2019/" + old_file_name.split("/")[-1]
        # keep track of things so we know what to copy later
        new_to_old_file_name[new_file_name] = old_file_name
        new_image["file_name"] = new_file_name
        new_test_metadata["images"].append(new_image)

    with open(public_path / "test2019.json", "w") as f:
        # The public test data, of course, doesn't have annotations
        public_new_test = new_test_metadata.copy()
        del public_new_test["annotations"]
        assert public_new_test.keys() == {
            "images",
            "info",
            "licenses",
        }, f"Public test metadata keys should be 'images', 'info', 'licenses', but found {public_new_test.keys()}"
        json.dump(public_new_test, f, indent=4, sort_keys=True)

    (public_path / "train_val2019").mkdir(parents=True, exist_ok=True)
    (public_path / "test2019").mkdir(parents=True, exist_ok=True)

    # Save private test answers
    answers_rows = []
    for image_info, annotation_info in zip(
        new_test_metadata["images"], new_test_metadata["annotations"]
    ):
        assert (
            image_info["id"] == annotation_info["image_id"]
        ), f"Mismatching image_id in image and annotation: {image_info['id']} vs {annotation_info['image_id']}"
        answers_rows.append(
            {
                "id": image_info["id"],
                "predicted": annotation_info["category_id"],
            }
        )
    answers_df = pd.DataFrame(answers_rows)
    answers_df.to_csv(private_path / "answers.csv", index=False)

    # Create new sample submission based on answers_df
    sample_df = answers_df.copy()
    sample_df["predicted"] = [random.Random(42).randint(0, 1009) for _ in range(len(sample_df))]
    sample_df.to_csv(public_path / "kaggle_sample_submission.csv", index=False)

    # Copy train images
    for annotation_image_metadata in tqdm(
        train_annotation_image_metadata_by_category.values(),
        desc=f"[{public_path.name}] Copying train images",
    ):
        for idx, annotation_image in enumerate(annotation_image_metadata):
            if dev_mode and idx >= image_count:
                break
            old_path = raw_path / annotation_image["image"]["file_name"]
            new_path = public_path / annotation_image["image"]["file_name"]
            new_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(old_path, new_path)

    # Copy test images
    for image_info in tqdm(
        new_test_metadata["images"], desc=f"[{public_path.name}] Copying test images"
    ):
        if dev_mode and len(new_to_old_file_name) >= image_count:
            break
        old_path = raw_path / new_to_old_file_name[image_info["file_name"]]
        new_path = public_path / image_info["file_name"]
        new_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(old_path, new_path)

    # Re-compress images
    add_to_tar(
        src=public_path / "test2019", out=public_path / "test2019.tar.gz"
    )  # Add to tar but don't actually compress with gzip to save time
    add_to_tar(src=public_path / "train_val2019", out=public_path / "train_val2019.tar.gz")
    # Remove uncompressed directories
    shutil.rmtree(public_path / "test2019")
    shutil.rmtree(public_path / "train_val2019")
    logger.info(f"Finished generating files for {public_path.name}")


def prepare(raw: Path, public: Path, private: Path):

    dev_mode = False
    image_count = 2 if dev_mode else float("inf")  # We copy over 2 images per category for dev mode

    # Extract train_val2019.tar.gz which contains images
    train_tar_path = raw / "train_val2019.tar.gz"
    train_extract_path = raw
    if not (raw / "train_val2019").exists():
        logger.info("Extracting raw image data...")
        shutil.unpack_archive(train_tar_path, train_extract_path)

    # Create train, test from train split
    json_path = raw / "train2019.json"
    with open(json_path, "r", encoding="utf-8") as f:
        old_train_metadata = json.load(f)

    # Organize data by category so that we can split per-category later
    annotation_image_metadata_by_category = {}  # We'll collect both `annotations` and `images` here
    for annotation_info, image_info in list(
        zip(old_train_metadata["annotations"], old_train_metadata["images"])
    ):
        assert (
            annotation_info["image_id"] == image_info["id"]
        ), f"Mismatching image_id in annotation and image: {annotation_info['image_id']} vs {image_info['id']}"
        category_id = annotation_info["category_id"]
        if category_id not in annotation_image_metadata_by_category:
            annotation_image_metadata_by_category[category_id] = []
        annotation_image_metadata_by_category[category_id].append(
            {
                "annotation": annotation_info,
                "image": image_info,
            }
        )

    # --- 1. Original Data Split (Train/Test) ---
    logger.info("--- Generating Original Train/Test Split ---")
    # Original train+val has 268,243 images, test has 35,400 images, ~0.12 ratio
    original_test_size = 0.12
    (
        original_train_split,
        original_test_split,
    ) = _split_by_category(
        annotation_image_metadata_by_category,
        test_size=original_test_size,
        random_state=0,
    )

    _generate_split_files(
        original_train_split,
        original_test_split,
        old_train_metadata,
        raw,
        public,
        private,
        dev_mode,
        image_count,
    )
    logger.info(f"Original split saved to {public.name} and {private.name}")

    # --- 2. New Validation Data Split (Train/Val) ---
    logger.info("--- Generating New Train/Validation Split ---")
    # Define new output directories
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # We want the new validation set ('test_val') to have the same size as the
    # original test set. We are splitting the *original_train_split* to get it.
    # test_val_size = new_test_size * train_size
    # We want: test_val_size ≈ test_size
    # So: new_test_size * (N * (1-0.12)) ≈ N * 0.12
    # new_test_size ≈ 0.12 / (1 - 0.12)
    val_split_test_size = original_test_size / (1.0 - original_test_size)

    (
        validation_train_split,
        validation_test_split,
    ) = _split_by_category(
        original_train_split,  # Split the TRAIN set from the first split
        test_size=val_split_test_size,
        random_state=0,  # Use same random state for consistency
    )

    _generate_split_files(
        validation_train_split,
        validation_test_split,
        old_train_metadata,
        raw,
        public_val,
        private_val,
        dev_mode,
        image_count,
    )
    logger.info(f"Validation split saved to {public_val.name} and {private_val.name}")


import shutil
from pathlib import Path

from sklearn.model_selection import train_test_split
from tqdm.auto import tqdm

from mlebench.utils import read_csv


def _create_split(
    df_to_split,
    raw_images_base_path: Path,
    public_path: Path,
    private_path: Path,
    test_size: float,
    random_state: int,
):
    """
    Helper function to perform a data split, create corresponding files, and copy images.

    Args:
        df_to_split: The DataFrame to be split into train and test sets.
        raw_images_base_path: Path to the directory containing the original images.
        public_path: The public output directory for this split.
        private_path: The private output directory for this split.
        test_size: The proportion of the dataset to allocate to the test split.
        random_state: The seed used by the random number generator.
    """
    # Create output directories
    public_path.mkdir(exist_ok=True, parents=True)
    private_path.mkdir(exist_ok=True, parents=True)

    # Perform the split
    new_train, new_test = train_test_split(
        df_to_split, test_size=test_size, random_state=random_state
    )
    sample_submission = new_test.copy()[["image", "hotel_id"]]
    sample_submission["hotel_id"] = "36363 53586 18807 64314 60181"

    # Save public files
    new_train.to_csv(public_path / "train.csv", index=False)
    sample_submission.to_csv(public_path / "sample_submission.csv", index=False)

    # Copy images from raw to prepared/public accordingly
    (public_path / "train_images").mkdir(exist_ok=True, parents=True)
    for image, chain in tqdm(
        zip(new_train["image"], new_train["chain"]),
        total=len(new_train),
        desc=f"Train images for {public_path.name}",
    ):
        chain = str(chain)
        (public_path / "train_images" / chain).mkdir(exist_ok=True, parents=True)
        shutil.copy(
            raw_images_base_path / chain / image,
            public_path / "train_images" / chain / image,
        )

    (public_path / "test_images").mkdir(exist_ok=True, parents=True)
    for image, chain in tqdm(
        zip(new_test["image"], new_test["chain"]),
        total=len(new_test),
        desc=f"Test images for {public_path.name}",
    ):
        chain = str(chain)
        # Note: Test images are copied to a flat directory structure
        shutil.copy(raw_images_base_path / chain / image, public_path / "test_images" / image)

    # Save private files
    new_test.to_csv(private_path / "test.csv", index=False)

    # Checks
    assert len(new_train) + len(new_test) == len(
        df_to_split
    ), "Length of new_train and new_test should equal length of input dataframe"
    assert sample_submission.columns.to_list() == [
        "image",
        "hotel_id",
    ], "Sample submission columns should only be `image` and `hotel_id`"
    assert len(sample_submission) == len(new_test), "Sample submission length should match test set"
    for image, chain in zip(new_train["image"], new_train["chain"]):
        chain = str(chain)
        assert (
            public_path / "train_images" / chain / image
        ).exists(), f"Image {image} not found in train_images folder"
    for image in new_test["image"]:
        assert (
            public_path / "test_images" / image
        ).exists(), f"Image {image} not found in test_images folder"
    assert not set(new_train["image"]).intersection(
        set(new_test["image"])
    ), "Train and test ids overlap"

    return new_train, new_test


def prepare(raw: Path, public: Path, private: Path):
    old_train = read_csv(raw / "train.csv")
    # drop image ce27d36c9147cc19.jpg: it appears twice and may occur across train and test when split
    old_train = old_train[old_train["image"] != "ce27d36c9147cc19.jpg"]

    # --- First Split (Original Public/Private) ---
    # This split produces the main benchmark data. The outputs in `public` and `private`
    # will be identical to the original script's output.
    original_train_df, _ = _create_split(
        df_to_split=old_train,
        raw_images_base_path=raw / "train_images",
        public_path=public,
        private_path=private,
        test_size=0.1,
        random_state=0,
    )

    # --- Second Split (New Validation Set) ---
    # This split takes the training data from the first split (`original_train_df`) and
    # splits it again to create a new, smaller training set and a validation set.
    # The results are saved in parallel `public_val` and `private_val` directories.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # The test size is 1/9 because we want the new validation set to be roughly the
    # same size as the original test set (10% of the total data).
    # original_train_df is 90% of the total, so 1/9 of it is 10% of the total.
    test_size_for_val_split = 1 / 9

    _create_split(
        df_to_split=original_train_df,
        raw_images_base_path=raw / "train_images",
        public_path=public_val,
        private_path=private_val,
        test_size=test_size_for_val_split,
        random_state=0,  # Use same random state for consistency and determinism
    )


import shutil
from pathlib import Path

import numpy as np
from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path):
    # Create train and test splits from train set
    old_train = read_csv(raw / "train_labels.csv", dtype={"BraTS21ID": str, "MGMT_value": int})
    new_train, new_test = train_test_split(old_train, test_size=0.1, random_state=0)

    # Copy over images
    (public / "train").mkdir(exist_ok=True)
    for file_id in new_train["BraTS21ID"]:
        (public / "train" / file_id).mkdir(exist_ok=True)
        shutil.copytree(
            src=raw / "train" / file_id,
            dst=public / "train" / file_id,
            dirs_exist_ok=True,
        )
    assert len(list(public.glob("train/*"))) == len(
        new_train
    ), "Public train should have the same number of images as the train set"

    (public / "test").mkdir(exist_ok=True)
    for file_id in new_test["BraTS21ID"]:
        (public / "test" / file_id).mkdir(exist_ok=True)
        shutil.copytree(
            src=raw / "train" / file_id,
            dst=public / "test" / file_id,
            dirs_exist_ok=True,
        )
    assert len(list(public.glob("test/*"))) == len(
        new_test
    ), "Public train should have the same number of images as the train set"

    # Create a sample submission file
    submission_df = new_test.copy()
    submission_df["MGMT_value"] = 0.5

    # Copy over files
    new_train.to_csv(public / "train_labels.csv", index=False)
    new_test.to_csv(private / "test.csv", index=False)
    submission_df.to_csv(public / "sample_submission.csv", index=False)



import shutil
from pathlib import Path

import py7zr
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import extract, read_csv


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    """
    # extract only what we need
    extract(raw / "train.7z", raw)
    extract(raw / "train_labels.csv.zip", raw)

    # Create train, test from train split
    # Original ratio is 1531/(1531+2295) = 0.4
    test_ratio = 0.2
    old_train = read_csv(raw / "train_labels.csv")
    new_train, answers = train_test_split(old_train, test_size=test_ratio, random_state=0)

    # Sample submission
    sample_submission = answers.copy()
    sample_submission["invasive"] = 0.5

    # Checks
    assert new_train["name"].is_unique, "new_train should have unique names"
    assert answers["name"].is_unique, "answers should have unique names"
    assert set(new_train["name"]).isdisjoint(
        set(answers["name"])
    ), "new_train and answers should be disjoint"
    assert len(new_train) + len(answers) == len(
        old_train
    ), "new_train and answers together should have the same number of rows as old_train"
    assert (
        new_train.columns.tolist() == old_train.columns.tolist()
    ), "new_train should have the same columns as old_train"
    assert (
        answers.columns.tolist() == old_train.columns.tolist()
    ), "answers should have the same columns as old_train"
    assert (
        sample_submission.columns.tolist() == old_train.columns.tolist()
    ), "sample_submission should have the same columns as old_train"

    # Write CSVs
    answers.to_csv(private / "answers.csv", index=False)
    new_train.to_csv(public / "train_labels.csv", index=False)
    sample_submission.to_csv(private / "sample_submission.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    # Copy files
    (public / "train").mkdir(exist_ok=True)
    (public / "test").mkdir(exist_ok=True)

    for file_id in tqdm(new_train["name"], desc="Copying Train Images"):
        shutil.copyfile(
            src=raw / "train" / f"{file_id}.jpg",
            dst=public / "train" / f"{file_id}.jpg",
        )

    for file_id in tqdm(answers["name"], desc="Copying Test Images"):
        shutil.copyfile(
            src=raw / "train" / f"{file_id}.jpg",
            dst=public / "test" / f"{file_id}.jpg",
        )

    # Checks
    assert len(list((public / "train").glob("*.jpg"))) == len(
        new_train
    ), "public/train should have the same number of files as new_train"
    assert len(list((public / "test").glob("*.jpg"))) == len(
        answers
    ), "public/test should have the same number of files as answers"

    # Zip
    shutil.make_archive(
        str(public / "sample_submission.csv"),
        "zip",
        root_dir=public,
        base_dir="sample_submission.csv",
    )
    shutil.make_archive(
        str(public / "train_labels.csv"), "zip", root_dir=public, base_dir="train_labels.csv"
    )
    with py7zr.SevenZipFile(public / "train.7z", "w") as z:
        z.write(public / "train")
    with py7zr.SevenZipFile(public / "test.7z", "w") as z:
        z.write(public / "test")

    # Delete
    shutil.rmtree(public / "train")
    shutil.rmtree(public / "test")
    (public / "sample_submission.csv").unlink()
    (public / "train_labels.csv").unlink()



"""
Helper functions from the Kaggle notebook `GSDC2 - baseline submission`.
Adapted from https://www.kaggle.com/code/saitodevel01/gsdc2-baseline-submission.
"""

from dataclasses import dataclass

import numpy as np

WGS84_SEMI_MAJOR_AXIS = 6378137.0
WGS84_SEMI_MINOR_AXIS = 6356752.314245
WGS84_SQUARED_FIRST_ECCENTRICITY = 6.69437999013e-3
WGS84_SQUARED_SECOND_ECCENTRICITY = 6.73949674226e-3

HAVERSINE_RADIUS = 6_371_000


@dataclass
class ECEF:
    x: np.array
    y: np.array
    z: np.array

    def to_numpy(self):
        return np.stack([self.x, self.y, self.z], axis=0)

    @staticmethod
    def from_numpy(pos):
        x, y, z = [np.squeeze(w) for w in np.split(pos, 3, axis=-1)]
        return ECEF(x=x, y=y, z=z)


@dataclass
class BLH:
    lat: np.array
    lng: np.array
    hgt: np.array


def ECEF_to_BLH(ecef):
    a = WGS84_SEMI_MAJOR_AXIS
    b = WGS84_SEMI_MINOR_AXIS
    e2 = WGS84_SQUARED_FIRST_ECCENTRICITY
    e2_ = WGS84_SQUARED_SECOND_ECCENTRICITY
    x = ecef.x
    y = ecef.y
    z = ecef.z
    r = np.sqrt(x**2 + y**2)
    t = np.arctan2(z * (a / b), r)
    B = np.arctan2(z + (e2_ * b) * np.sin(t) ** 3, r - (e2 * a) * np.cos(t) ** 3)
    L = np.arctan2(y, x)
    n = a / np.sqrt(1 - e2 * np.sin(B) ** 2)
    H = (r / np.cos(B)) - n

    return BLH(lat=B, lng=L, hgt=H)


def haversine_distance(blh_1, blh_2):
    dlat = blh_2.lat - blh_1.lat
    dlng = blh_2.lng - blh_1.lng
    a = np.sin(dlat / 2) ** 2 + np.cos(blh_1.lat) * np.cos(blh_2.lat) * np.sin(dlng / 2) ** 2
    dist = 2 * HAVERSINE_RADIUS * np.arcsin(np.sqrt(a))

    return dist


def pandas_haversine_distance(df1, df2):
    blh1 = BLH(
        lat=np.deg2rad(df1["LatitudeDegrees"].to_numpy()),
        lng=np.deg2rad(df1["LongitudeDegrees"].to_numpy()),
        hgt=0,
    )

    blh2 = BLH(
        lat=np.deg2rad(df2["LatitudeDegrees"].to_numpy()),
        lng=np.deg2rad(df2["LongitudeDegrees"].to_numpy()),
        hgt=0,
    )

    return haversine_distance(blh1, blh2)


def calc_score(pred_df, gt_df):
    d = pandas_haversine_distance(pred_df, gt_df)
    score = np.mean([np.quantile(d, 0.50), np.quantile(d, 0.95)])
    return score



import os
import shutil
import zipfile
from pathlib import Path

import numpy as np
from tqdm.auto import tqdm

from mlebench.utils import extract, get_logger

np_rng = np.random.RandomState(0)

logger = get_logger(__name__)


def count_lines_in_file(file_path):
    line_count = 0
    with open(file_path, "r", encoding="utf-8") as file:
        for _line in file:
            line_count += 1
    return line_count


def compress_file_to_zip(src_file: Path, zip_file: Path):
    with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zipf:
        zipf.write(src_file, arcname=src_file.name)


def remove_random_word(sentence: str) -> str:
    """
    Remove a random 'word' (sequence of characters, delimited by whitespace) from a sentence.
    Does not remove first or last words.

    Punctuation counts as a word, and is already separated by whitespace.
    """
    words = sentence.split()
    index = np_rng.randint(1, len(words) - 1)
    return " ".join(words[:index] + words[index + 1 :])


def _split_and_process_data(
    input_file: Path,
    output_train_file: Path,
    output_public_test_file: Path,
    output_private_test_file: Path,
    total_lines: int,
) -> tuple[int, int]:
    """
    Helper function to perform the core data splitting and processing logic.

    Reads from an input file and splits it into a train and test set based on a
    probabilistic condition, writing them to the specified output files.
    """
    with (
        open(input_file, "r", encoding="utf-8") as old_train,
        open(output_train_file, "w", encoding="utf-8") as public_train,
        open(output_public_test_file, "w", encoding="utf-8") as public_test,
        open(output_private_test_file, "w", encoding="utf-8") as private_test,
    ):
        public_test.write('"id","sentence"\n')
        private_test.write('"id","sentence"\n')
        test_count = 0
        train_count = 0
        # there is one sentence per line
        for sentence in tqdm(old_train, desc=f"Processing {input_file.name}", total=total_lines):
            # we will put ~0.01 of the data in test, the rest in train, matching kaggle's original split
            # some sentences only have 2 words, so can't remove a word -- keep them in train
            if np_rng.uniform() <= 0.01 and len(sentence.strip().split()) > 2:
                # get rid of linebreak and escape quotes
                sentence_clean = sentence.strip().replace('"', '""')
                removed_word_sentence = remove_random_word(sentence_clean)
                private_test.write(f'{test_count},"{sentence_clean}"\n')
                public_test.write(f'{test_count},"{removed_word_sentence}"\n')
                test_count += 1
            else:
                public_train.write(sentence)
                train_count += 1
    return train_count, test_count


def prepare(raw: Path, public: Path, private: Path):
    logger.info("Extracting raw / train_v2.txt.zip")
    extract(raw / "train_v2.txt.zip", raw)

    # Define and create the new validation directories
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"
    public_val.mkdir(exist_ok=True, parents=True)
    private_val.mkdir(exist_ok=True, parents=True)

    # --- 1. Original Split (raw -> train/test) ---
    logger.info("--- Generating original train/test split ---")
    # computed this ahead of time
    total_lines = 30301028
    original_train_count, original_test_count = _split_and_process_data(
        input_file=raw / "train_v2.txt",
        output_train_file=public / "train_v2.txt",
        output_public_test_file=public / "test_v2.txt",
        output_private_test_file=private / "test.csv",
        total_lines=total_lines,
    )
    assert (
        original_train_count + original_test_count == total_lines
    ), "Sum of train and test samples must equal total samples for original split."

    # --- 2. Second Split (train -> train_val/test_val) ---
    logger.info("--- Generating validation split from the new training set ---")
    # The input for the second split is the training set from the first split.
    val_split_input_file = public / "train_v2.txt"
    val_split_total_lines = original_train_count
    val_train_count, val_test_count = _split_and_process_data(
        input_file=val_split_input_file,
        output_train_file=public_val / "train_v2.txt",
        output_public_test_file=public_val / "test_v2.txt",
        output_private_test_file=private_val / "test.csv",
        total_lines=val_split_total_lines,
    )
    assert (
        val_train_count + val_test_count == val_split_total_lines
    ), "Sum of train_val and test_val samples must equal total samples for validation split."

    # --- 3. Process and Compress Original public/private directories ---
    logger.info("--- Compressing and cleaning up original public/private directories ---")
    # we will be compressing the public files (to match what's on kaggle.com)
    # so copy our sample submission to private so we have access to it
    shutil.copy(public / "test_v2.txt", private / "sample_submission.csv")

    # compress the public files
    logger.info("Compressing train_v2.txt")
    compress_file_to_zip(public / "train_v2.txt", public / "train_v2.txt.zip")
    logger.info("Compressing test_v2.txt")
    compress_file_to_zip(public / "test_v2.txt", public / "test_v2.txt.zip")
    # remove the original files
    (public / "train_v2.txt").unlink()
    (public / "test_v2.txt").unlink()

    # --- 4. Process and Compress New public_val/private_val directories ---
    logger.info("--- Compressing and cleaning up validation public_val/private_val directories ---")
    # Replicate the process for the validation set
    shutil.copy(public_val / "test_v2.txt", private_val / "sample_submission.csv")

    # compress the public_val files
    logger.info("Compressing validation train_v2.txt")
    compress_file_to_zip(public_val / "train_v2.txt", public_val / "train_v2.txt.zip")
    logger.info("Compressing validation test_v2.txt")
    compress_file_to_zip(public_val / "test_v2.txt", public_val / "test_v2.txt.zip")
    # remove the original files
    (public_val / "train_v2.txt").unlink()
    (public_val / "test_v2.txt").unlink()

    # --- 5. Final Checks ---
    logger.info("--- Running final checks ---")
    # Original Checks
    assert not (public / "train_v2.txt").exists(), "public / 'train_v2.txt' should not exist"
    assert (public / "train_v2.txt.zip").exists(), "public / 'train_v2.txt.zip' should exist"
    assert not (public / "test_v2.txt").exists(), "public / 'test_v2.txt' should not exist"
    assert (public / "test_v2.txt.zip").exists(), "public / 'test_v2.txt.zip' should exist"

    private_test_line_count = count_lines_in_file(private / "test.csv")
    assert (
        private_test_line_count - 1 == original_test_count
    ), "private / 'test.csv' has incorrect number of lines"
    assert (
        count_lines_in_file(private / "sample_submission.csv") == private_test_line_count
    ), "private / 'sample_submission.csv' has incorrect number of lines"

    # New Checks for Validation Set
    assert not (public_val / "train_v2.txt").exists(), "public_val / 'train_v2.txt' should not exist"
    assert (public_val / "train_v2.txt.zip").exists(), "public_val / 'train_v2.txt.zip' should exist"
    assert not (public_val / "test_v2.txt").exists(), "public_val / 'test_v2.txt' should not exist"
    assert (public_val / "test_v2.txt.zip").exists(), "public_val / 'test_v2.txt.zip' should exist"

    private_val_test_line_count = count_lines_in_file(private_val / "test.csv")
    assert (
        private_val_test_line_count - 1 == val_test_count
    ), "private_val / 'test.csv' has incorrect number of lines"
    assert (
        count_lines_in_file(private_val / "sample_submission.csv") == private_val_test_line_count
    ), "private_val / 'sample_submission.csv' has incorrect number of lines"


import logging
import re
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

from dsat.benchmark.benchmark import BaseBenchmark
from dsat.services.llm import LLMService
from dsat.services.sandbox import SandboxService
from dsat.services.workspace import WorkspaceService
from dsat.utils.context import MAX_HISTORY_CHARS, MAX_OUTPUT_CHARS, truncate_output
from dsat.workflows.base import DSATWorkflow

logger = logging.getLogger(__name__)


class DeepAnalyzeWorkflow(DSATWorkflow):
    """
    Workflow implementation for DeepAnalyze-8B.

    DeepAnalyze uses structured tags (<Analyze>, <Code>, <Execute>, <Answer>) and requires
    multi-round dialog where the system injects code execution results into <Execute> tags.
    """

    # Common output file extensions to look for
    OUTPUT_EXTENSIONS = {'.csv', '.txt', '.json', '.xlsx', '.xls', '.png', '.jpg', '.jpeg', '.pdf', '.html', '.py', '.pkl', '.pickle', '.npy', '.npz', '.h5', '.hdf5', '.parquet'}
    
    # Files to ignore when scanning for outputs
    IGNORE_FILES = {'prompt.json', '.gitkeep', '.DS_Store', 'thumbs.db'}

    def __init__(
        self,
        operators: Dict[str, Any],
        services: Dict[str, Any],
        agent_config: Dict[str, Any],
        benchmark: Optional[BaseBenchmark] = None,
    ):
        super().__init__(operators, services, agent_config)
        self.llm_service: LLMService = services["llm"]
        self.sandbox_service: SandboxService = services["sandbox"]
        self.workspace_service: WorkspaceService = services.get("workspace")
        self.execute_op = operators.get("execute")
        if not self.execute_op:
            raise ValueError("DeepAnalyzeWorkflow requires an 'execute' operator.")

        self.max_iterations = agent_config.get("max_iterations", 10)
        self.benchmark = benchmark

    @staticmethod
    def _extract_code_from_segment(segment: str) -> Optional[str]:
        """Extract python code between <Code>...</Code>, optionally fenced by ```python ... ```."""
        code_match = re.search(r"<Code>(.*?)</Code>", segment, re.DOTALL)
        if not code_match:
            return None
        code_content = code_match.group(1).strip()
        md_match = re.search(r"```(?:python)?(.*?)```", code_content, re.DOTALL)
        return (md_match.group(1).strip() if md_match else code_content)

    @staticmethod
    def _extract_answer_from_segment(segment: str) -> Optional[str]:
        """Extract answer content between <Answer>...</Answer> tags."""
        # Try case-insensitive match for <Answer> or <answer>
        answer_match = re.search(r"<Answer>(.*?)</Answer>", segment, re.DOTALL | re.IGNORECASE)
        if not answer_match:
            return None
        answer_content = answer_match.group(1).strip()
        # Remove "Answer:" prefix if present
        if "Answer:" in answer_content:
            answer_content = answer_content.split("Answer:", 1)[-1].strip()
        return answer_content

    def _should_terminate(self, response: str, iteration: int) -> bool:
        """Stop when an <Answer> block appears or max iterations reached."""
        # Some models emit a full <Answer>...</Answer> in one turn; others may emit
        # a bare </Answer> later. Prefer terminating as soon as any Answer block
        # is detected to avoid an extra empty turn.
        if re.search(r"<Answer>.*?</Answer>", response, re.DOTALL | re.IGNORECASE) or "</Answer>" in response:
            logger.info("Detected <Answer> block, stopping iterations.")
            return True
        if iteration + 1 >= self.max_iterations:
            logger.info("Reached maximum iterations, stopping workflow.")
            return True
        return False

    def _build_llm_messages(self, conversation_history: List[Dict[str, str]]) -> List[Dict[str, str]]:
        if not conversation_history:
            return []

        max_chars = MAX_HISTORY_CHARS
        first = conversation_history[0]
        first_content = truncate_output(first.get("content", ""), max_chars // 2)
        messages = [{"role": first.get("role", "user"), "content": first_content}]
        total_chars = len(first_content)

        recent_messages = []
        for msg in reversed(conversation_history[1:]):
            content = truncate_output(msg.get("content", ""), MAX_OUTPUT_CHARS)
            msg_len = len(content)
            if total_chars + msg_len > max_chars:
                break
            recent_messages.append({"role": msg.get("role", "user"), "content": content})
            total_chars += msg_len

        messages.extend(reversed(recent_messages))
        return messages

    @staticmethod
    def _extract_output_filenames_from_description(description: str) -> List[str]:
        """
        Extract all potential output filenames from task description.
        Returns a list of filenames (without path) that might be expected outputs.
        """
        filenames = []
        
        # Patterns to match various ways output files are specified
        patterns = [
            # "saved in/to 'filename.ext'" or "saved in/to \"filename.ext\""
            r'saved?\s+(?:in|to|as)\s+["\']([^"\']+\.\w+)["\']',
            # "output file named 'filename.ext'"
            r'output\s+file\s+(?:named|called)?\s*["\']([^"\']+\.\w+)["\']',
            # "save the results/output to 'filename.ext'"
            r'save\s+(?:the\s+)?(?:results?|output|data|file)\s+(?:to|in|as)\s+["\']([^"\']+\.\w+)["\']',
            # "results should be saved in 'filename.ext'"
            r'(?:results?|output)\s+should\s+be\s+saved\s+(?:in|to|as)\s+["\']([^"\']+\.\w+)["\']',
            # "write to 'filename.ext'"
            r'write\s+(?:to|into)\s+["\']([^"\']+\.\w+)["\']',
            # "export to 'filename.ext'"
            r'export\s+(?:to|as)\s+["\']([^"\']+\.\w+)["\']',
            # "filename.csv" or 'filename.csv' standalone in quotes (common patterns)
            r'["\']([a-zA-Z0-9_\-]+\.(?:csv|txt|json|xlsx|png|jpg|pdf|html|py))["\']',
        ]
        
        seen = set()
        for pattern in patterns:
            for match in re.finditer(pattern, description, re.IGNORECASE):
                filename = match.group(1)
                # Clean up the filename
                filename = filename.strip()
                # Only add if it looks like a valid filename and not seen before
                if filename and '/' not in filename and '\\' not in filename and filename not in seen:
                    filenames.append(filename)
                    seen.add(filename)
        
        return filenames

    def _get_initial_sandbox_files(self, sandbox_workdir: Path) -> Set[str]:
        """Get the set of files initially present in sandbox (to detect new files later)."""
        if not sandbox_workdir.exists():
            return set()
        return {f.name for f in sandbox_workdir.iterdir() if f.is_file()}

    def _find_new_output_files(self, sandbox_workdir: Path, initial_files: Set[str]) -> List[Path]:
        """
        Find all new files created in sandbox since initial state.
        Returns list of Path objects for new output files.
        """
        new_files = []
        if not sandbox_workdir.exists():
            return new_files
            
        for f in sandbox_workdir.iterdir():
            if not f.is_file():
                continue
            if f.name in initial_files:
                continue
            if f.name.lower() in self.IGNORE_FILES:
                continue
            if f.name.startswith('_sandbox_script_'):
                continue
            # Check if it's a recognized output type
            if f.suffix.lower() in self.OUTPUT_EXTENSIONS or f.suffix == '':
                new_files.append(f)
                
        return new_files

    def _collect_outputs_to_destination(
        self,
        sandbox_workdir: Path,
        output_path: Path,
        expected_filenames: List[str],
        initial_files: Set[str],
    ) -> bool:
        """
        Collect output files from sandbox to destination directory.
        
        Strategy:
        1. First, check if the expected output file (output_path.name) exists
        2. Then, check for any files matching expected_filenames from task description
        3. Finally, collect any new files created during execution
        
        All matching files are copied to output_path.parent, preserving original names.
        The primary output is also copied to output_path for compatibility.
        
        Returns True if at least one output file was collected.
        """
        output_path.parent.mkdir(parents=True, exist_ok=True)
        collected = False
        copied_files = set()
        
        # 1. Check for the default expected output file
        default_output = sandbox_workdir / output_path.name
        if default_output.exists():
            shutil.copy(default_output, output_path)
            copied_files.add(output_path.name)
            collected = True
            logger.info("Copied default output file to %s", output_path)
        
        # 2. Check for files specified in task description
        for filename in expected_filenames:
            src_file = sandbox_workdir / filename
            if src_file.exists() and filename not in copied_files:
                dst_file = output_path.parent / filename
                shutil.copy(src_file, dst_file)
                copied_files.add(filename)
                collected = True
                logger.info("Copied expected output file '%s' to %s", filename, dst_file)
                
                # If no default output was found, also copy first expected file as the default
                if not (output_path.exists()):
                    shutil.copy(src_file, output_path)
                    logger.info("Also copied '%s' as default output to %s", filename, output_path)
        
        # 3. Collect any other new files created during execution
        new_files = self._find_new_output_files(sandbox_workdir, initial_files)
        for src_file in new_files:
            if src_file.name not in copied_files:
                dst_file = output_path.parent / src_file.name
                shutil.copy(src_file, dst_file)
                copied_files.add(src_file.name)
                collected = True
                logger.info("Copied new output file '%s' to %s", src_file.name, dst_file)
                
                # If still no default output, use first new file
                if not output_path.exists():
                    shutil.copy(src_file, output_path)
                    logger.info("Also copied '%s' as default output to %s", src_file.name, output_path)
        
        if collected:
            logger.info("Total %d output file(s) collected to %s", len(copied_files), output_path.parent)
        
        return collected

    def _write_answer_to_file(self, output_path: Path, answer_content: str) -> None:
        """Write answer content to output file based on file extension."""
        output_path.parent.mkdir(parents=True, exist_ok=True)
        
        suffix = output_path.suffix.lower()
        
        if suffix == '.txt':
            output_path.write_text(answer_content, encoding='utf-8')
            logger.info("Answer written to text file: %s", output_path)
            
        elif suffix == '.csv':
            # For CSV files, check if content looks like CSV data
            if '\n' in answer_content and ',' in answer_content:
                output_path.write_text(answer_content, encoding='utf-8')
            else:
                # Wrap simple answer in CSV format
                output_path.write_text(f"answer\n{answer_content}\n", encoding='utf-8')
            logger.info("Answer written to CSV file: %s", output_path)
            
        elif suffix == '.json':
            # Try to format as JSON if possible
            import json
            try:
                # Check if it's already valid JSON
                json.loads(answer_content)
                output_path.write_text(answer_content, encoding='utf-8')
            except json.JSONDecodeError:
                # Wrap in JSON format
                output_path.write_text(json.dumps({"answer": answer_content}), encoding='utf-8')
            logger.info("Answer written to JSON file: %s", output_path)
            
        else:
            # Default: write as plain text
            output_path.write_text(answer_content, encoding='utf-8')
            logger.info("Answer written to file: %s", output_path)

    @staticmethod
    def _build_initial_prompt(
        description: str,
        io_instructions: str,
        data_dir: Path,
        output_path: Path,
    ) -> str:
        return f"""You are an expert data scientist using the DeepAnalyze model to solve data science tasks.

Task Description:
{description}

Input/Output Requirements:
{io_instructions}

Data Directory: {data_dir}
Output File: {output_path}

Please use the following format to analyze and solve the problem:
1. Analyze the task in the <Analyze>...</Analyze> tags
2. Provide the code to execute in the <Code>...</Code> tags (may include ```python``` code blocks)
3. The system will provide code execution results in the <Execute>...</Execute> tags
4. When you determine that the code execution results can successfully solve the problem, output the final answer in the <Answer>...</Answer> tags

Please begin the first round of analysis."""

    async def solve(
        self,
        description: str,
        io_instructions: str,
        data_dir: Path,
        output_path: Path,
    ) -> None:
        """
        Execute the DeepAnalyze workflow to solve a data science task.
        
        The workflow:
        1. Sends task description to LLM
        2. Extracts and executes code from <Code> blocks
        3. Feeds execution results back via <Execute> blocks
        4. Repeats until <Answer> tag is found or max iterations reached
        5. Collects all output files from sandbox to destination
        """
        logger.info(
            "DeepAnalyzeWorkflow starting to solve task. Target output: %s",
            output_path,
        )

        if not self.workspace_service:
            raise ValueError("WorkspaceService is required for DeepAnalyzeWorkflow.")

        # Extract expected output filenames from task description
        expected_filenames = self._extract_output_filenames_from_description(description)
        if expected_filenames:
            logger.info("Expected output files from task description: %s", expected_filenames)

        # Get sandbox workdir and record initial files
        sandbox_workdir = self.workspace_service.get_path("sandbox_workdir")
        initial_files = self._get_initial_sandbox_files(sandbox_workdir)
        logger.debug("Initial sandbox files: %s", initial_files)

        initial_prompt = self._build_initial_prompt(
            description=description,
            io_instructions=io_instructions,
            data_dir=data_dir,
            output_path=output_path,
        )
        conversation_history: List[Dict[str, str]] = [
            {"role": "user", "content": initial_prompt}
        ]

        final_answer_content: Optional[str] = None
        # Keep the most recent assistant segment that contained an <Answer> block,
        # since some models emit a bare </Answer> in a follow-up turn.
        last_answer_segment: Optional[str] = None
        # If the model fails to emit a <Code> block, retry within the same iteration
        # with a short nudge, up to this many times.
        no_code_retries_max: int = int(self.agent_config.get("no_code_retries", 2))

        for iteration in range(self.max_iterations):
            logger.info(
                "--- DeepAnalyze Iteration %d/%d ---",
                iteration + 1,
                self.max_iterations,
            )

            no_code_retry_count = 0
            extracted_code: Optional[str] = None
            llm_response = ""

            while True:
                # LLMService.call() only supports single prompts, so use the underlying multi-turn helper.
                response = await self.llm_service._make_llm_call_with_retries(  # type: ignore[attr-defined]
                    messages=self._build_llm_messages(conversation_history),
                    max_retries=self.llm_service.config.max_retries,
                )
                llm_response = response.choices[0].message.content
                logger.debug("DeepAnalyze raw response: %s", llm_response)
                conversation_history.append({"role": "assistant", "content": llm_response})

                # Extract answer if present
                extracted_answer = self._extract_answer_from_segment(llm_response)
                if extracted_answer:
                    final_answer_content = extracted_answer
                    last_answer_segment = llm_response
                    logger.info("Extracted answer from <Answer> tag: %s", extracted_answer[:100] + "..." if len(extracted_answer) > 100 else extracted_answer)

                if self._should_terminate(llm_response, iteration):
                    break

                extracted_code = self._extract_code_from_segment(llm_response)
                if extracted_code:
                    break

                if no_code_retry_count >= no_code_retries_max:
                    logger.warning(
                        "No <Code> block found after %d retry(ies) in iteration %d, continuing to next iteration.",
                        no_code_retry_count,
                        iteration + 1,
                    )
                    break

                no_code_retry_count += 1
                logger.warning(
                    "No <Code> block found in iteration %d, retrying (%d/%d).",
                    iteration + 1,
                    no_code_retry_count,
                    no_code_retries_max,
                )
                conversation_history.append(
                    {
                        "role": "user",
                        "content": (
                            "<Feedback>\n"
                            "No <Code>...</Code> block was detected. "
                            "Please provide executable Python code inside <Code> tags so it can be run.\n"
                            "</Feedback>\n"
                        ),
                    }
                )

            if self._should_terminate(llm_response, iteration):
                break

            if not extracted_code:
                continue

            exec_result = await self.execute_op(code=extracted_code, mode="script")
            if exec_result.success:
                logger.info("Code execution succeeded in iteration %d.", iteration + 1)
                execution_output = exec_result.stdout or "Code executed successfully."
            else:
                logger.warning("Code execution failed in iteration %d.", iteration + 1)
                execution_output = f"Error during execution:\n{exec_result.stderr}"

            safe_execution_output = truncate_output(execution_output, MAX_OUTPUT_CHARS)
            execute_block = f"\n<Execute>\n{safe_execution_output}\n</Execute>\n"
            conversation_history.append({"role": "user", "content": execute_block})

            # After successful execution, collect intermediate outputs
            if exec_result.success:
                collected = self._collect_outputs_to_destination(
                    sandbox_workdir=sandbox_workdir,
                    output_path=output_path,
                    expected_filenames=expected_filenames,
                    initial_files=initial_files,
                )
                if collected:
                    logger.info("Intermediate outputs collected in iteration %d", iteration + 1)

        # Final output collection after all iterations
        logger.info("Workflow iterations complete. Performing final output collection...")
        collected = self._collect_outputs_to_destination(
            sandbox_workdir=sandbox_workdir,
            output_path=output_path,
            expected_filenames=expected_filenames,
            initial_files=initial_files,
        )

        if not collected:
            # No files were collected from sandbox
            if not final_answer_content:
                # Try to recover answer from the last assistant segment that had it.
                if last_answer_segment:
                    final_answer_content = self._extract_answer_from_segment(last_answer_segment)
                # Fallback: scan conversation history for the most recent <Answer>.
                if not final_answer_content:
                    for msg in reversed(conversation_history):
                        if msg.get("role") != "assistant":
                            continue
                        recovered = self._extract_answer_from_segment(msg.get("content", ""))
                        if recovered:
                            final_answer_content = recovered
                            break

            if final_answer_content:
                logger.info(
                    "No output files found in sandbox, but answer content recovered from <Answer>. Writing to output file."
                )
                self._write_answer_to_file(output_path, final_answer_content)
            else:
                logger.warning(
                    "DeepAnalyzeWorkflow finished but no output files found in sandbox and no <Answer> tag content extracted."
                )
        
        logger.info("DeepAnalyzeWorkflow completed. Output directory: %s", output_path.parent)



from pandas import DataFrame
from sklearn.metrics import mean_absolute_error

from mlebench.grade_helpers import InvalidSubmissionError


def prepare_for_grading(submission: DataFrame, answers: DataFrame, target_column: str) -> DataFrame:
    """
    Merge the submission's target_column into the answers DataFrame, matching on 'segment_id'.
    target_column from the submission DataFrame will have the suffix '_pred',
    and target_column from the answers DataFrame will have the suffix '_true'.
    """

    # Answers checks
    assert (
        target_column in answers.columns
    ), f"Target column {target_column} not found in answers DataFrame."
    assert "segment_id" in answers.columns, "Segment ID column not found in answers DataFrame."

    # Submission checks
    if target_column not in submission.columns:
        raise InvalidSubmissionError(
            f"Target column {target_column} not found in submissions DataFrame."
        )
    if "segment_id" not in submission.columns:
        raise InvalidSubmissionError("Segment ID column not found in submission DataFrame.")
    if not set(submission["segment_id"]) == set(answers["segment_id"]):
        raise InvalidSubmissionError(
            f"Submission is missing the following segment_ids: {set(answers['segment_id']) - set(submission['segment_id'])}"
        )

    merged = answers.merge(
        submission[["segment_id", target_column]], on="segment_id", suffixes=("_true", "_pred")
    )
    return merged


def grade(submission: DataFrame, answers: DataFrame) -> float:
    merged = prepare_for_grading(submission, answers, "time_to_eruption")
    score = mean_absolute_error(
        y_true=merged["time_to_eruption_true"], y_pred=merged["time_to_eruption_pred"]
    )
    rounded_score = int(round(score))
    return rounded_score



# dsat/services/states/journal.py

"""
Implements JournalState, which manages a tree of solution attempts (Nodes).
This is the core state representation for Paradigm 2 (AIDE/AutoMind-style) search agents.
"""
import uuid
from functools import total_ordering
from typing import Optional, Any, List, Dict, Set

from pydantic import BaseModel, Field, ConfigDict

from dsat.common.typing import ExecutionResult
from dsat.utils.context import truncate_output
from dsat.services.states.base import State


@total_ordering
class MetricValue(BaseModel):
    """
    Represents a comparable metric that can be configured for maximization or minimization.
    A value of None is considered worse than any numeric value.
    """
    value: Optional[float]
    maximize: bool = True

    def __gt__(self, other: "MetricValue") -> bool:
        if self.value is None:
            return False
        if other.value is None:
            return True
        return (self.value > other.value) if self.maximize else (self.value < other.value)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, MetricValue) and self.value == other.value

    def __str__(self) -> str:
        direction = "↑" if self.maximize else "↓"
        val_str = f"{self.value:.4f}" if self.value is not None else "N/A"
        return f"Metric{direction}({val_str})"

class Node(BaseModel):
    """
    Represents a single attempt or node in the solution search tree.
    Each node contains the code, plan, execution results, and review analysis.
    """
    code: str
    plan: str

    id: str = Field(default_factory=lambda: uuid.uuid4().hex)
    parent_id: Optional[str] = None
    children_ids: Set[str] = Field(default_factory=set)

    # Execution Results
    term_out: str = ""
    exec_time: float = 0.0
    exc_type: Optional[str] = None
    exec_metadata: Dict[str, Any] = Field(default_factory=dict)

    # LLM Recordings
    task_context: Dict[str, Any] = Field(default_factory=dict)
    generate_prompt: Optional[str] = None
    llm_generate: Optional[Dict[str, Any]] = None
    review_context: Optional[Dict[str, Any]] = None
    llm_review: Optional[Dict[str, Any]] = None

    # Review Results
    analysis: str = ""
    metric: MetricValue = Field(default_factory=lambda: MetricValue(value=None))
    is_buggy: bool = True
    step: int = -1

    # Artifact paths
    code_artifact_path: Optional[str] = None
    final_submission_path: Optional[str] = None

    def absorb_exec_result(self, exec_result: ExecutionResult):
        """Updates the node with the results from a sandbox execution."""
        stdout = exec_result.stdout or ""
        stderr = exec_result.stderr or ""
        combined_output = f"STDOUT:\n{stdout}\n\nSTDERR:\n{stderr}".strip()
        self.term_out = truncate_output(combined_output)
        self.exc_type = exec_result.exc_type
        self.is_buggy = not exec_result.success
        self.exec_metadata = exec_result.metadata or {}

    model_config = ConfigDict(
        # Note: json_encoders deprecated in Pydantic V2
        # Sets are now automatically serialized to lists
    )

class JournalState(State, BaseModel):
    """
    Manages the entire search tree (the "Journal") of solution nodes.
    Provides methods for appending nodes, traversing the tree, and selecting
    nodes based on different criteria (e.g., best, buggy).
    """
    nodes: Dict[str, Node] = Field(default_factory=dict)

    def __len__(self) -> int:
        return len(self.nodes)

    def append(self, node: Node, parent: Optional[Node] = None):
        """Adds a new node to the journal, linking it to a parent if provided."""
        if parent:
            if parent.id not in self.nodes:
                 raise ValueError(f"Parent node with id {parent.id} not in journal.")
            node.parent_id = parent.id
            self.nodes[parent.id].children_ids.add(node.id)
        node.step = len(self)
        self.nodes[node.id] = node

    def get_node(self, node_id: str) -> Optional[Node]:
        """Retrieves a node by its ID."""
        return self.nodes.get(node_id)

    def get_best_node(self) -> Optional[Node]:
        """Finds the best-performing, non-buggy node in the entire journal."""
        good_nodes = [n for n in self.nodes.values() if not n.is_buggy]
        if not good_nodes:
            return None
        return max(good_nodes, key=lambda n: n.metric)

    def generate_summary(self, max_nodes: int = 3) -> str:
        """
        Creates a textual summary of successful past attempts for prompt context.
        MODIFIED: Now selects the `max_nodes` BEST performing successful attempts.
        """
        good_nodes = sorted(
            [n for n in self.nodes.values() if not n.is_buggy and n.metric.value is not None],
            key=lambda x: x.metric,
            reverse=True  # MetricValue handles > comparison correctly, so reverse=True gets the best
        )
        if not good_nodes:
            return "No successful solutions have been found yet."

        # Apply windowing: take the top `max_nodes`
        selected_nodes = good_nodes[:max_nodes]

        summary_parts = []
        for n in selected_nodes:
            summary_part = (
                f"Attempt #{n.step}:\n"
                f"Plan: {n.plan}\n"
                f"Result Analysis: {n.analysis}\n"
                f"Validation Metric: {n.metric}\n"
            )
            summary_parts.append(summary_part)
        
        prefix = ""
        if len(good_nodes) > len(selected_nodes):
            prefix = f"[... {len(good_nodes) - len(selected_nodes)} other successful attempts exist ...]\n"

        return prefix + "Here is a summary of the best performing attempts:\n" + "\n------------------\n".join(summary_parts)



import shutil
from pathlib import Path

from pandas import DataFrame, read_csv
from sklearn.model_selection import train_test_split

from mlebench.utils import extract


def prepare(raw: Path, public: Path, private: Path):
    """
    Prepares the dataset by performing two sequential splits.
    1. Splits the raw data into a main train/test set for the `public` and
       `private` directories.
    2. Splits the main training set again to create a smaller train/validation
       set for the `public_val` and `private_val` directories.
    """

    def _split_and_save(
        data_to_split: DataFrame,
        test_ratio: float,
        public_dir: Path,
        private_dir: Path,
        random_state: int,
    ) -> DataFrame:
        """
        Helper function to perform a data split, save files to specified
        directories, and return the resulting training set for a potential
        subsequent split.
        """
        # Ensure output directories exist
        public_dir.mkdir(parents=True, exist_ok=True)
        private_dir.mkdir(parents=True, exist_ok=True)

        # Create train and test splits from the provided dataframe
        new_train, answers = train_test_split(
            data_to_split, test_size=test_ratio, random_state=random_state
        )

        # Create public test set (unlabeled)
        new_test = answers.copy()
        new_test = new_test.drop("Sentiment", axis="columns")

        # Create sample submission
        sample_submission = answers[["PhraseId", "Sentiment"]].copy()
        sample_submission["Sentiment"] = 2

        # Checks
        assert new_train["PhraseId"].is_unique, f"PhraseId in new_train ({public_dir.name}) should be unique"
        assert new_test["PhraseId"].is_unique, f"PhraseId in new_test ({public_dir.name}) should be unique"
        assert set(new_train["PhraseId"]).isdisjoint(
            set(new_test["PhraseId"])
        ), f"PhraseId in new_train and new_test ({public_dir.name}) should be disjoint"
        assert (
            new_train.shape[0] + new_test.shape[0] == data_to_split.shape[0]
        ), "New train and new test should have the same number of rows as the input data"
        assert (
            new_train.columns.tolist() == data_to_split.columns.tolist()
        ), "New train and input data should have the same columns"
        assert new_test.columns.tolist() == [
            "PhraseId",
            "SentenceId",
            "Phrase",
        ], "new_test should have columns ['PhraseId', 'SentenceId', 'Phrase']"

        # Write CSVs to their respective directories
        answers.to_csv(private_dir / "answers.csv", index=False)
        new_train.to_csv(public_dir / "train.tsv", index=False, sep="\t")
        new_test.to_csv(public_dir / "test.tsv", index=False, sep="\t")
        sample_submission.to_csv(public_dir / "sampleSubmission.csv", index=False)

        # Zip files
        shutil.make_archive(str(public_dir / "train.tsv"), "zip", public_dir, "train.tsv")
        shutil.make_archive(str(public_dir / "test.tsv"), "zip", public_dir, "test.tsv")

        # Delete unzipped files
        (public_dir / "train.tsv").unlink()
        (public_dir / "test.tsv").unlink()

        return new_train

    # --- Main script execution starts here ---

    # Extract the raw data file
    extract(raw / "train.tsv.zip", raw)
    old_train = read_csv(raw / "train.tsv", sep="\t")

    # Define the test ratio for the first split
    test_ratio_1 = 0.3  # 66293/(156061+66293) = 0.3

    # --- First Split: Create the original public/private sets ---
    # This call produces the original, unmodified competition output.
    # The resulting training set is captured for the second split.
    train_for_val_split = _split_and_save(
        data_to_split=old_train,
        test_ratio=test_ratio_1,
        public_dir=public,
        private_dir=private,
        random_state=0,
    )

    # --- Second Split: Create the new validation sets ---
    # Define the new parallel directories for the validation split
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Calculate the test ratio for the second split. This ensures the new
    # validation set (`test_val`) is the same size as the original test set.
    # test_ratio_2 = (size of original test set) / (size of new training set)
    #              = (total * test_ratio_1) / (total * (1 - test_ratio_1))
    test_ratio_2 = test_ratio_1 / (1 - test_ratio_1)

    # Perform the second split on the training data from the first split
    _split_and_save(
        data_to_split=train_for_val_split,
        test_ratio=test_ratio_2,
        public_dir=public_val,
        private_dir=private_val,
        random_state=0,
    )


from pathlib import Path

from sklearn.model_selection import train_test_split

from mlebench.competitions.utils import df_to_one_hot
from mlebench.utils import extract, read_csv

from .classes import CLASSES


def prepare(raw: Path, public: Path, private: Path):
    target_col = "author"
    id_col = "id"

    # extract only what we need
    extract(raw / "train.zip", raw)

    # Create train, test from train split
    old_train = read_csv(raw / "train.csv")
    new_train, new_test = train_test_split(old_train, test_size=0.1, random_state=0)
    new_test_without_labels = new_test.drop(columns=[target_col])

    # private test matches the format of sample submission
    one_hot_new_test = df_to_one_hot(
        new_test.drop(columns=["text"]),
        id_column=id_col,
        target_column=target_col,
        classes=CLASSES,
    )
    # fill the sample submission with arbitrary values (matching kaggle.com)
    sample_submission = one_hot_new_test.copy()
    sample_submission["EAP"] = 0.403493538995863
    sample_submission["HPL"] = 0.287808366106543
    sample_submission["MWS"] = 0.308698094897594

    # save files
    new_train.to_csv(public / "train.csv", index=False)
    new_test_without_labels.to_csv(public / "test.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)
    one_hot_new_test.to_csv(private / "test.csv", index=False)
